@@ -68,11 +68,28 @@ rdns_send_request (struct rdns_request *req, int fd, bool new_req)
 	}
 
 	if (resolver->curve_plugin == NULL) {
-		r = send (fd, req->packet, req->pos, 0);
+		if (!req->io->connected) {
+			r = sendto (fd, req->packet, req->pos, 0,
+					req->io->saddr,
+					req->io->slen);
+		}
+		else {
+			r = send (fd, req->packet, req->pos, 0);
+		}
 	}
 	else {
-		r = resolver->curve_plugin->cb.curve_plugin.send_cb (req,
-				resolver->curve_plugin->data);
+		if (!req->io->connected) {
+			r = resolver->curve_plugin->cb.curve_plugin.send_cb (req,
+					resolver->curve_plugin->data,
+					req->io->saddr,
+					req->io->slen);
+		}
+		else {
+			r = resolver->curve_plugin->cb.curve_plugin.send_cb (req,
+					resolver->curve_plugin->data,
+					NULL,
+					0);
+		}
 	}
 	if (r == -1) {
 		if (errno == EAGAIN || errno == EINTR) {
@@ -81,26 +98,39 @@ rdns_send_request (struct rdns_request *req, int fd, bool new_req)
 				HASH_ADD_INT (req->io->requests, id, req);
 				req->async_event = resolver->async->add_write (resolver->async->data,
 					fd, req);
+				req->state = RDNS_REQUEST_WAIT_SEND;
 			}
 			/*
 			 * If request is already processed then the calling function
 			 * should take care about events processing
 			 */
 			return 0;
-		} 
+		}
 		else {
 			rdns_debug ("send failed: %s for server %s", strerror (errno), serv->name);
 			return -1;
 		}
 	}
-	
+	else if (!req->io->connected) {
+		/* Connect socket */
+		r = connect (fd, req->io->saddr, req->io->slen);
+
+		if (r == -1) {
+			rdns_err ("cannot connect after sending request: %s for server %s",
+					strerror (errno), serv->name);
+		}
+		else {
+			req->io->connected = true;
+		}
+	}
+
 	if (new_req) {
 		/* Add request to hash table */
 		HASH_ADD_INT (req->io->requests, id, req);
 		/* Fill timeout */
 		req->async_event = resolver->async->add_timer (resolver->async->data,
 				req->timeout, req);
-		req->state = RDNS_REQUEST_SENT;
+		req->state = RDNS_REQUEST_WAIT_REPLY;
 	}
 
 	return 1;
@@ -119,6 +149,7 @@ rdns_make_reply (struct rdns_request *req, enum dns_rcode rcode)
 		rep->entries = NULL;
 		rep->code = rcode;
 		req->reply = rep;
+		rep->authenticated = false;
 	}
 
 	return rep;
@@ -131,7 +162,7 @@ rdns_find_dns_request (uint8_t *in, struct rdns_io_channel *ioc)
 	struct rdns_request *req;
 	int id;
 	struct rdns_resolver *resolver = ioc->resolver;
-	
+
 	id = header->qid;
 	HASH_FIND_INT (ioc->requests, &id, req);
 	if (req == NULL) {
@@ -170,7 +201,7 @@ rdns_parse_reply (uint8_t *in, int r, struct rdns_request *req,
 		return false;
 	}
 
-	/* 
+	/*
 	 * Now we have request and query data is now at the end of header, so compare
 	 * request QR section and reply QR section
 	 */
@@ -190,6 +221,10 @@ rdns_parse_reply (uint8_t *in, int r, struct rdns_request *req,
 	 */
 	rep = rdns_make_reply (req, header->rcode);
 
+	if (header->ad) {
+		rep->authenticated = true;
+	}
+
 	if (rep == NULL) {
 		rdns_warn ("Cannot allocate memory for reply");
 		return false;
@@ -221,25 +256,18 @@ rdns_parse_reply (uint8_t *in, int r, struct rdns_request *req,
 			}
 		}
 	}
-	
+
 	if (!found && type != RDNS_REQUEST_ANY) {
 		/* We have not found the requested RR type */
-		rep->code = RDNS_RC_NOREC;
+		if (rep->code == RDNS_RC_NOERROR) {
+			rep->code = RDNS_RC_NOREC;
+		}
 	}
 
 	*_rep = rep;
 	return true;
 }
 
-static void
-rdns_request_unschedule (struct rdns_request *req)
-{
-	req->async->del_timer (req->async->data,
-			req->async_event);
-	/* Remove from id hashes */
-	HASH_DEL (req->io->requests, req);
-}
-
 void
 rdns_process_read (int fd, void *arg)
 {
@@ -251,17 +279,18 @@ rdns_process_read (int fd, void *arg)
 	uint8_t in[UDP_PACKET_SIZE];
 
 	resolver = ioc->resolver;
-	
+
 	/* First read packet from socket */
 	if (resolver->curve_plugin == NULL) {
-		r = read (fd, in, sizeof (in));
+		r = recv (fd, in, sizeof (in), 0);
 		if (r > (int)(sizeof (struct dns_header) + sizeof (struct dns_query))) {
 			req = rdns_find_dns_request (in, ioc);
 		}
 	}
 	else {
 		r = resolver->curve_plugin->cb.curve_plugin.recv_cb (ioc, in,
-				sizeof (in), resolver->curve_plugin->data, &req);
+				sizeof (in), resolver->curve_plugin->data, &req,
+				ioc->saddr, ioc->slen);
 		if (req == NULL &&
 				r > (int)(sizeof (struct dns_header) + sizeof (struct dns_query))) {
 			req = rdns_find_dns_request (in, ioc);
@@ -271,8 +300,14 @@ rdns_process_read (int fd, void *arg)
 	if (req != NULL) {
 		if (rdns_parse_reply (in, r, req, &rep)) {
 			UPSTREAM_OK (req->io->srv);
-			req->state = RDNS_REQUEST_REPLIED;
+
+			if (req->resolver->ups && req->io->srv->ups_elt) {
+				req->resolver->ups->ok (req->io->srv->ups_elt,
+						req->resolver->ups->data);
+			}
+
 			rdns_request_unschedule (req);
+			req->state = RDNS_REQUEST_REPLIED;
 			req->func (rep, req->arg);
 			REF_RELEASE (req);
 		}
@@ -292,63 +327,124 @@ rdns_process_timer (void *arg)
 	bool renew = false;
 	struct rdns_resolver *resolver;
 	struct rdns_server *serv = NULL;
+	unsigned cnt;
 
 	req->retransmits --;
 	resolver = req->resolver;
 
 	if (req->retransmits == 0) {
-		UPSTREAM_FAIL (req->io->srv, time (NULL));
+		if (req->resolver->ups && req->io->srv->ups_elt) {
+			req->resolver->ups->fail (req->io->srv->ups_elt,
+					req->resolver->ups->data);
+		}
+		else {
+			UPSTREAM_FAIL (req->io->srv, time (NULL));
+		}
+
 		rep = rdns_make_reply (req, RDNS_RC_TIMEOUT);
-		req->state = RDNS_REQUEST_REPLIED;
 		rdns_request_unschedule (req);
+		req->state = RDNS_REQUEST_REPLIED;
 		req->func (rep, req->arg);
 		REF_RELEASE (req);
 
 		return;
 	}
 
-	if (!req->io->active) {
-		/* Do not reschedule IO requests on inactive sockets */
-		rdns_debug ("reschedule request with id: %d", (int)req->id);
-		rdns_request_unschedule (req);
-		REF_RELEASE (req->io);
+	if (!req->io->active || req->retransmits == 1) {
 
-		UPSTREAM_SELECT_ROUND_ROBIN (resolver->servers, serv);
-
-		if (serv == NULL) {
-			rdns_warn ("cannot find suitable server for request");
-			rep = rdns_make_reply (req, RDNS_RC_SERVFAIL);
-			req->state = RDNS_REQUEST_REPLIED;
-			req->func (rep, req->arg);
-			REF_RELEASE (req);
+		if (resolver->ups) {
+			cnt = resolver->ups->count (resolver->ups->data);
+		}
+		else {
+			cnt = 0;
+			UPSTREAM_FOREACH (resolver->servers, serv) {
+				cnt ++;
+			}
 		}
 
-		/* Select random IO channel */
-		req->io = serv->io_channels[ottery_rand_uint32 () % serv->io_cnt];
-		req->io->uses ++;
-		REF_RETAIN (req->io);
-		renew = true;
+		if (!req->io->active || cnt > 1) {
+			/* Do not reschedule IO requests on inactive sockets */
+			rdns_debug ("reschedule request with id: %d", (int)req->id);
+			rdns_request_unschedule (req);
+			REF_RELEASE (req->io);
+
+			if (resolver->ups) {
+				struct rdns_upstream_elt *elt;
+
+				elt = resolver->ups->select_retransmit (req->requested_names[0].name,
+						req->requested_names[0].len, resolver->ups->data);
+
+				if (elt) {
+					serv = elt->server;
+					serv->ups_elt = elt;
+				}
+				else {
+					UPSTREAM_SELECT_ROUND_ROBIN (resolver->servers, serv);
+				}
+			}
+			else {
+				UPSTREAM_SELECT_ROUND_ROBIN (resolver->servers, serv);
+			}
+
+			if (serv == NULL) {
+				rdns_warn ("cannot find suitable server for request");
+				rep = rdns_make_reply (req, RDNS_RC_SERVFAIL);
+				req->state = RDNS_REQUEST_REPLIED;
+				req->func (rep, req->arg);
+				REF_RELEASE (req);
+
+				return;
+			}
+
+			/* Select random IO channel */
+			req->io = serv->io_channels[ottery_rand_uint32 () % serv->io_cnt];
+			req->io->uses ++;
+			REF_RETAIN (req->io);
+			renew = true;
+		}
 	}
 
+	/*
+	 * Note: when `renew` is true, then send_request deals with the
+	 * timers and events itself
+	 */
 	r = rdns_send_request (req, req->io->sock, renew);
 	if (r == 0) {
 		/* Retransmit one more time */
-		req->async->del_timer (req->async->data,
+		if (!renew) {
+			req->async->del_timer (req->async->data,
 					req->async_event);
-		req->async_event = req->async->add_write (req->async->data,
-				req->io->sock, req);
-		req->state = RDNS_REQUEST_REGISTERED;
+			req->async_event = req->async->add_write (req->async->data,
+					req->io->sock, req);
+		}
+
+		req->state = RDNS_REQUEST_WAIT_SEND;
 	}
 	else if (r == -1) {
-		UPSTREAM_FAIL (req->io->srv, time (NULL));
+		if (req->resolver->ups && req->io->srv->ups_elt) {
+			req->resolver->ups->fail (req->io->srv->ups_elt,
+					req->resolver->ups->data);
+		}
+		else {
+			UPSTREAM_FAIL (req->io->srv, time (NULL));
+		}
+
+		if (!renew) {
+			req->async->del_timer (req->async->data,
+					req->async_event);
+			req->async_event = NULL;
+			HASH_DEL (req->io->requests, req);
+		}
+
+		/* We have not scheduled timeout actually due to send error */
 		rep = rdns_make_reply (req, RDNS_RC_NETERR);
 		req->state = RDNS_REQUEST_REPLIED;
-		rdns_request_unschedule (req);
 		req->func (rep, req->arg);
 		REF_RELEASE (req);
 	}
 	else {
 		req->async->repeat_timer (req->async->data, req->async_event);
+		req->state = RDNS_REQUEST_WAIT_REPLY;
 	}
 }
 
@@ -380,7 +476,7 @@ rdns_process_ioc_refresh (void *arg)
 						continue;
 					}
 					nioc->sock = rdns_make_client_socket (serv->name, serv->port,
-							SOCK_DGRAM);
+							SOCK_DGRAM, &nioc->saddr, &nioc->slen);
 					if (nioc->sock == -1) {
 						rdns_err ("cannot open socket to %s: %s", serv->name,
 								strerror (errno));
@@ -416,6 +512,15 @@ rdns_process_retransmit (int fd, void *arg)
 
 	resolver->async->del_write (resolver->async->data,
 			req->async_event);
+	req->async_event = NULL;
+
+	if (req->state == RDNS_REQUEST_FAKE) {
+		/* Reply is ready */
+		req->func (req->reply, req->arg);
+		REF_RELEASE (req);
+
+		return;
+	}
 
 	r = rdns_send_request (req, fd, false);
 
@@ -423,10 +528,17 @@ rdns_process_retransmit (int fd, void *arg)
 		/* Retransmit one more time */
 		req->async_event = req->async->add_write (req->async->data,
 						fd, req);
-		req->state = RDNS_REQUEST_REGISTERED;
+		req->state = RDNS_REQUEST_WAIT_SEND;
 	}
 	else if (r == -1) {
-		UPSTREAM_FAIL (req->io->srv, time (NULL));
+		if (req->resolver->ups && req->io->srv->ups_elt) {
+			req->resolver->ups->fail (req->io->srv->ups_elt,
+					req->resolver->ups->data);
+		}
+		else {
+			UPSTREAM_FAIL (req->io->srv, time (NULL));
+		}
+
 		rep = rdns_make_reply (req, RDNS_RC_NETERR);
 		req->state = RDNS_REQUEST_REPLIED;
 		req->func (rep, req->arg);
@@ -435,10 +547,13 @@ rdns_process_retransmit (int fd, void *arg)
 	else {
 		req->async_event = req->async->add_timer (req->async->data,
 			req->timeout, req);
-		req->state = RDNS_REQUEST_SENT;
+		req->state = RDNS_REQUEST_WAIT_REPLY;
 	}
 }
 
+#define align_ptr(p, a)                                                   \
+    (guint8 *) (((uintptr_t) (p) + ((uintptr_t) a - 1)) & ~((uintptr_t) a - 1))
+
 struct rdns_request*
 rdns_make_request_full (
 		struct rdns_resolver *resolver,
@@ -458,13 +573,24 @@ rdns_make_request_full (
 	size_t olen;
 	const char *cur_name, *last_name = NULL;
 	struct rdns_compression_entry *comp = NULL;
+	struct rdns_fake_reply *fake_rep = NULL;
+	char fake_buf[MAX_FAKE_NAME + sizeof (struct rdns_fake_reply_idx) + 16];
+	struct rdns_fake_reply_idx *idx;
+
+	if (resolver == NULL || !resolver->initialized) {
+		if (resolver == NULL) {
+			return NULL;
+		}
+
+		rdns_err ("resolver is uninitialized");
 
-	if (!resolver->initialized) {
 		return NULL;
 	}
 
 	req = malloc (sizeof (struct rdns_request));
 	if (req == NULL) {
+		rdns_err ("failed to allocate memory for request: %s",
+				strerror (errno));
 		return NULL;
 	}
 
@@ -477,8 +603,13 @@ rdns_make_request_full (
 	req->state = RDNS_REQUEST_NEW;
 	req->packet = NULL;
 	req->requested_names = calloc (queries, sizeof (struct rdns_request_name));
+	req->async_event = NULL;
+
 	if (req->requested_names == NULL) {
 		free (req);
+		rdns_err ("failed to allocate memory for request data: %s",
+				strerror (errno));
+
 		return NULL;
 	}
 
@@ -487,91 +618,182 @@ rdns_make_request_full (
 	req->curve_plugin_data = NULL;
 #endif
 	REF_INIT_RETAIN (req, rdns_request_free);
-	
+
 	/* Calculate packet's total length based on records count */
 	va_start (args, queries);
 	for (i = 0; i < queries * 2; i += 2) {
 		cur = i / 2;
 		cur_name = va_arg (args, const char *);
+		type = va_arg (args, int);
+
 		if (cur_name != NULL) {
-			last_name = cur_name;
 			clen = strlen (cur_name);
+
 			if (clen == 0) {
-				rdns_info ("got empty name to resolve");
+				rdns_warn ("got empty name to resolve");
 				rdns_request_free (req);
 				return NULL;
 			}
+
+			if (cur_name[0] == '.') {
+				/* Skip dots at the begin */
+				unsigned int ndots = strspn (cur_name, ".");
+
+				cur_name += ndots;
+				clen -= ndots;
+
+				if (clen == 0) {
+					rdns_warn ("got empty name to resolve");
+					rdns_request_free (req);
+					return NULL;
+				}
+			}
+
+			if (cur_name[clen - 1] == '.') {
+				/* Skip trailing dots */
+				while (clen >= 1 && cur_name[clen - 1] == '.') {
+					clen --;
+				}
+
+				if (clen == 0) {
+					rdns_warn ("got empty name to resolve");
+					rdns_request_free (req);
+					return NULL;
+				}
+			}
+
+			if (last_name == NULL && queries == 1 && clen < MAX_FAKE_NAME) {
+				/* We allocate structure in the static space */
+				idx = (struct rdns_fake_reply_idx *)align_ptr (fake_buf, 16);
+				idx->type = type;
+				idx->len = clen;
+				memcpy (idx->request, cur_name, clen);
+				HASH_FIND (hh, resolver->fake_elts, idx, sizeof (*idx) + clen,
+						fake_rep);
+
+				if (fake_rep) {
+					/* We actually treat it as a short-circuit */
+					req->reply = rdns_make_reply (req, fake_rep->rcode);
+					req->reply->entries = fake_rep->result;
+					req->state = RDNS_REQUEST_FAKE;
+				}
+			}
+
+			last_name = cur_name;
 			tlen += clen;
 		}
 		else if (last_name == NULL) {
-			rdns_info ("got NULL as the first name to resolve");
+			rdns_err ("got NULL as the first name to resolve");
 			rdns_request_free (req);
 			return NULL;
 		}
 
-		if (!rdns_format_dns_name (resolver, last_name, clen,
-				&req->requested_names[cur].name, &olen)) {
-			rdns_request_free (req);
-			return NULL;
-		}
-
-		type = va_arg (args, int);
-		req->requested_names[cur].type = type;
-		req->requested_names[cur].len = olen;
-	}
-	va_end (args);
-
-	rdns_allocate_packet (req, tlen);
-	rdns_make_dns_header (req, queries);
-
-	for (i = 0; i < queries; i ++) {
-		cur_name = req->requested_names[i].name;
-		clen = req->requested_names[i].len;
-		type = req->requested_names[i].type;
-		if (queries > 1) {
-			if (!rdns_add_rr (req, cur_name, clen, type, &comp)) {
-				REF_RELEASE (req);
-				rnds_compression_free (comp);
+		if (req->state != RDNS_REQUEST_FAKE) {
+			if (!rdns_format_dns_name (resolver, last_name, clen,
+					&req->requested_names[cur].name, &olen)) {
+				rdns_err ("cannot format %s", last_name);
+				rdns_request_free (req);
 				return NULL;
 			}
+
+			req->requested_names[cur].len = olen;
 		}
 		else {
-			if (!rdns_add_rr (req, cur_name, clen, type, NULL)) {
-				REF_RELEASE (req);
-				rnds_compression_free (comp);
-				return NULL;
-			}
+			req->requested_names[cur].len = clen;
 		}
+
+		req->requested_names[cur].type = type;
 	}
 
-	rnds_compression_free (comp);
+	va_end (args);
 
-	/* Add EDNS RR */
-	rdns_add_edns0 (req);
+	if (req->state != RDNS_REQUEST_FAKE) {
+		rdns_allocate_packet (req, tlen);
+		rdns_make_dns_header (req, queries);
+
+		for (i = 0; i < queries; i++) {
+			cur_name = req->requested_names[i].name;
+			clen = req->requested_names[i].len;
+			type = req->requested_names[i].type;
+			if (queries > 1) {
+				if (!rdns_add_rr (req, cur_name, clen, type, &comp)) {
+					rdns_err ("cannot add rr", cur_name);
+					REF_RELEASE (req);
+					rnds_compression_free (comp);
+					return NULL;
+				}
+			} else {
+				if (!rdns_add_rr (req, cur_name, clen, type, NULL)) {
+					rdns_err ("cannot add rr", cur_name);
+					REF_RELEASE (req);
+					rnds_compression_free (comp);
+					return NULL;
+				}
+			}
+		}
+
+		rnds_compression_free (comp);
+
+		/* Add EDNS RR */
+		rdns_add_edns0 (req);
+
+		req->retransmits = repeats;
+		req->timeout = timeout;
+		req->state = RDNS_REQUEST_NEW;
+	}
 
-	req->retransmits = repeats;
-	req->timeout = timeout;
-	req->state = RDNS_REQUEST_NEW;
 	req->async = resolver->async;
 
-	UPSTREAM_SELECT_ROUND_ROBIN (resolver->servers, serv);
+	if (resolver->ups) {
+		struct rdns_upstream_elt *elt;
+
+		elt = resolver->ups->select (req->requested_names[0].name,
+				req->requested_names[0].len, resolver->ups->data);
+
+		if (elt) {
+			serv = elt->server;
+			serv->ups_elt = elt;
+		}
+		else {
+			UPSTREAM_SELECT_ROUND_ROBIN (resolver->servers, serv);
+		}
+	}
+	else {
+		UPSTREAM_SELECT_ROUND_ROBIN (resolver->servers, serv);
+	}
 
 	if (serv == NULL) {
 		rdns_warn ("cannot find suitable server for request");
 		REF_RELEASE (req);
 		return NULL;
 	}
-	
+
 	/* Select random IO channel */
 	req->io = serv->io_channels[ottery_rand_uint32 () % serv->io_cnt];
-	req->io->uses ++;
-	
-	/* Now send request to server */
-	r = rdns_send_request (req, req->io->sock, true);
 
-	if (r == -1) {
-		REF_RELEASE (req);
-		return NULL;
+	if (req->state == RDNS_REQUEST_FAKE) {
+		req->async_event = resolver->async->add_write (resolver->async->data,
+				req->io->sock, req);
+	}
+	else {
+		req->io->uses++;
+
+		/* Now send request to server */
+		r = rdns_send_request (req, req->io->sock, true);
+
+		if (r == -1) {
+			rdns_info ("cannot send DNS request: %s", strerror (errno));
+			REF_RELEASE (req);
+
+			if (resolver->ups && serv->ups_elt) {
+				resolver->ups->fail (serv->ups_elt, resolver->ups->data);
+			}
+			else {
+				UPSTREAM_FAIL (serv, time (NULL));
+			}
+
+			return NULL;
+		}
 	}
 
 	REF_RETAIN (req->io);
@@ -588,10 +810,12 @@ rdns_resolver_init (struct rdns_resolver *resolver)
 	struct rdns_io_channel *ioc;
 
 	if (!resolver->async_binded) {
+		rdns_err ("no async backend specified");
 		return false;
 	}
-	
+
 	if (resolver->servers == NULL) {
+		rdns_err ("no DNS servers defined");
 		return false;
 	}
 
@@ -601,13 +825,17 @@ rdns_resolver_init (struct rdns_resolver *resolver)
 		for (i = 0; i < serv->io_cnt; i ++) {
 			ioc = calloc (1, sizeof (struct rdns_io_channel));
 			if (ioc == NULL) {
-				rdns_err ("cannot allocate memory for the resolver");
+				rdns_err ("cannot allocate memory for the resolver IO channels");
 				return false;
 			}
-			ioc->sock = rdns_make_client_socket (serv->name, serv->port, SOCK_DGRAM);
-			ioc->active = true;
+
+			ioc->sock = rdns_make_client_socket (serv->name, serv->port, SOCK_DGRAM,
+					&ioc->saddr, &ioc->slen);
+
 			if (ioc->sock == -1) {
-				rdns_err ("cannot open socket to %s:%d %s", serv->name, serv->port, strerror (errno));
+				ioc->active = false;
+				rdns_err ("cannot open socket to %s:%d %s",
+						serv->name, serv->port, strerror (errno));
 				free (ioc);
 				return false;
 			}
@@ -644,7 +872,7 @@ rdns_resolver_register_plugin (struct rdns_resolver *resolver,
 	}
 }
 
-bool
+void *
 rdns_resolver_add_server (struct rdns_resolver *resolver,
 		const char *name, unsigned int port,
 		int priority, unsigned int io_cnt)
@@ -658,24 +886,24 @@ rdns_resolver_add_server (struct rdns_resolver *resolver,
 	if (inet_pton (AF_INET, name, &addr) == 0 &&
 		inet_pton (AF_INET6, name, &addr) == 0) {
 		/* Invalid IP */
-		return false;
+		return NULL;
 	}
 
 	if (io_cnt == 0) {
-		return false;
+		return NULL;
 	}
 	if (port == 0 || port > UINT16_MAX) {
-		return false;
+		return NULL;
 	}
 
 	serv = calloc (1, sizeof (struct rdns_server));
 	if (serv == NULL) {
-		return false;
+		return NULL;
 	}
 	serv->name = strdup (name);
 	if (serv->name == NULL) {
 		free (serv);
-		return false;
+		return NULL;
 	}
 
 	serv->io_cnt = io_cnt;
@@ -683,7 +911,7 @@ rdns_resolver_add_server (struct rdns_resolver *resolver,
 
 	UPSTREAM_ADD (resolver->servers, serv, priority);
 
-	return true;
+	return serv;
 }
 
 void
@@ -701,6 +929,15 @@ rdns_resolver_set_log_level (struct rdns_resolver *resolver,
 	resolver->log_level = level;
 }
 
+void
+rdns_resolver_set_upstream_lib (struct rdns_resolver *resolver,
+		struct rdns_upstream_context *ups_ctx,
+		void *ups_data)
+{
+	resolver->ups = ups_ctx;
+	resolver->ups->data = ups_data;
+}
+
 
 void
 rdns_resolver_set_max_io_uses (struct rdns_resolver *resolver,
@@ -780,3 +1017,57 @@ rdns_resolver_async_bind (struct rdns_resolver *resolver,
 		resolver->async_binded = true;
 	}
 }
+
+void
+rdns_resolver_set_dnssec (struct rdns_resolver *resolver, bool enabled)
+{
+	if (resolver) {
+		resolver->enable_dnssec = enabled;
+	}
+}
+
+
+void rdns_resolver_set_fake_reply (struct rdns_resolver *resolver,
+								   const char *name,
+								   enum rdns_request_type type,
+								   enum dns_rcode rcode,
+								   struct rdns_reply_entry *reply)
+{
+	struct rdns_fake_reply *fake_rep;
+	struct rdns_fake_reply_idx *srch;
+	unsigned len = strlen (name);
+
+	assert (len < MAX_FAKE_NAME);
+	srch = malloc (sizeof (*srch) + len);
+	srch->len = len;
+	srch->type = type;
+	memcpy (srch->request, name, len);
+
+	HASH_FIND (hh, resolver->fake_elts, srch, len + sizeof (*srch), fake_rep);
+
+	if (fake_rep) {
+		/* Append reply to the existing list */
+		fake_rep->rcode = rcode;
+
+		if (reply) {
+			DL_CONCAT (fake_rep->result, reply);
+		}
+	}
+	else {
+		fake_rep = calloc (1, sizeof (*fake_rep) + len);
+
+		if (fake_rep == NULL) {
+			abort ();
+		}
+
+		fake_rep->rcode = rcode;
+
+		memcpy (&fake_rep->key, srch, sizeof (*srch) + len);
+
+		if (reply) {
+			DL_CONCAT (fake_rep->result, reply);
+		}
+
+		HASH_ADD (hh, resolver->fake_elts, key, sizeof (*srch) + len, fake_rep);
+	}
+}
\ No newline at end of file
