@@ -1,27 +1,18 @@
-/*
- * Copyright (c) 2015, Vsevolod Stakhov
- * All rights reserved.
+/*-
+ * Copyright 2016 Vsevolod Stakhov
  *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions are met:
- *	 * Redistributions of source code must retain the above copyright
- *	   notice, this list of conditions and the following disclaimer.
- *	 * Redistributions in binary form must reproduce the above copyright
- *	   notice, this list of conditions and the following disclaimer in the
- *	   documentation and/or other materials provided with the distribution.
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
  *
- * THIS SOFTWARE IS PROVIDED BY AUTHOR ''AS IS'' AND ANY
- * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
- * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
- * DISCLAIMED. IN NO EVENT SHALL AUTHOR BE LIABLE FOR ANY
- * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
- * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
- * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
- * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
- * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
- * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ *   http://www.apache.org/licenses/LICENSE-2.0
+ *
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
  */
-
 #include "config.h"
 #include "expression.h"
 #include "printf.h"
@@ -29,46 +20,32 @@
 #include "util.h"
 #include "utlist.h"
 #include "ottery.h"
+#include <math.h>
 
 #define RSPAMD_EXPR_FLAG_NEGATE (1 << 0)
 #define RSPAMD_EXPR_FLAG_PROCESSED (1 << 1)
 
 #define MIN_RESORT_EVALS 50
 #define MAX_RESORT_EVALS 150
+#define DOUBLE_EPSILON 1e-9
 
-enum rspamd_expression_op {
-	OP_INVALID = 0,
-	OP_PLUS, /* || or + */
-	OP_MULT, /* && or * */
-	OP_OR, /* || or | */
-	OP_AND, /* && or & */
-	OP_NOT, /* ! */
-	OP_LT, /* < */
-	OP_GT, /* > */
-	OP_LE, /* <= */
-	OP_GE, /* >= */
-	OP_OBRACE, /* ( */
-	OP_CBRACE /* ) */
+enum rspamd_expression_elt_type {
+	ELT_OP = 0,
+	ELT_ATOM,
+	ELT_LIMIT
 };
 
 struct rspamd_expression_elt {
-	enum {
-		ELT_OP = 0,
-		ELT_ATOM,
-		ELT_LIMIT
-	} type;
-
+	enum rspamd_expression_elt_type type;
 	union {
 		rspamd_expression_atom_t *atom;
 		enum rspamd_expression_op op;
-		struct {
-			gint val;
-			gint op_idx;
-		} lim;
+		gdouble lim;
 	} p;
+
 	gint flags;
-	gint value;
 	gint priority;
+	gdouble value;
 };
 
 struct rspamd_expression {
@@ -168,6 +145,23 @@ rspamd_expr_stack_pop (struct rspamd_expression *expr)
 	return rspamd_expr_stack_elt_pop (expr->expression_stack);
 }
 
+static gpointer
+rspamd_expr_stack_peek (struct rspamd_expression *expr)
+{
+	gpointer e;
+	gint idx;
+	GPtrArray *stack = expr->expression_stack;
+
+	if (stack->len == 0) {
+		return NULL;
+	}
+
+	idx = stack->len - 1;
+	e = g_ptr_array_index (stack, idx);
+
+	return e;
+}
+
 /*
  * Return operation priority
  */
@@ -360,9 +354,17 @@ rspamd_expression_destroy (struct rspamd_expression *expr)
 			}
 		}
 
-		g_array_free (expr->expressions, TRUE);
-		g_ptr_array_free (expr->expression_stack, TRUE);
-		g_node_destroy (expr->ast);
+		if (expr->expressions) {
+			g_array_free (expr->expressions, TRUE);
+		}
+		if (expr->expression_stack) {
+			g_ptr_array_free (expr->expression_stack, TRUE);
+		}
+		if (expr->ast) {
+			g_node_destroy (expr->ast);
+		}
+
+		g_free (expr);
 	}
 }
 
@@ -380,12 +382,21 @@ rspamd_ast_add_node (GPtrArray *operands, struct rspamd_expression_elt *op,
 		/* Unary operator */
 		res = g_node_new (op);
 		a1 = rspamd_expr_stack_elt_pop (operands);
-		g_node_append (res, a1);
+
 		if (a1 == NULL) {
 			g_set_error (err, rspamd_expr_quark(), EINVAL, "no operand to "
 					"unary '%s' operation", rspamd_expr_op_to_str (op->p.op));
+			g_node_destroy (res);
+
 			return FALSE;
 		}
+
+		g_node_append (res, a1);
+		test_elt = a1->data;
+
+		if (test_elt->type == ELT_ATOM) {
+			test_elt->p.atom->parent = res;
+		}
 	}
 	else {
 		/* For binary operators we might want to examine chains */
@@ -402,6 +413,7 @@ rspamd_ast_add_node (GPtrArray *operands, struct rspamd_expression_elt *op,
 					"'%s' operation", rspamd_expr_op_to_str (op->p.op));
 			return FALSE;
 		}
+
 		/* First try with a1 */
 		test = a1;
 		test_elt = test->data;
@@ -412,6 +424,7 @@ rspamd_ast_add_node (GPtrArray *operands, struct rspamd_expression_elt *op,
 			rspamd_expr_stack_elt_push (operands, a1);
 			return TRUE;
 		}
+
 		/* Now test a2 */
 		test = a2;
 		test_elt = test->data;
@@ -427,6 +440,16 @@ rspamd_ast_add_node (GPtrArray *operands, struct rspamd_expression_elt *op,
 		res = g_node_new (op);
 		g_node_append (res, a1);
 		g_node_append (res, a2);
+
+		test_elt = a1->data;
+		if (test_elt->type == ELT_ATOM) {
+			test_elt->p.atom->parent = res;
+		}
+
+		test_elt = a2->data;
+		if (test_elt->type == ELT_ATOM) {
+			test_elt->p.atom->parent = res;
+		}
 	}
 
 	/* Push back resulting node to the stack */
@@ -484,6 +507,13 @@ rspamd_ast_priority_cmp (GNode *a, GNode *b)
 	struct rspamd_expression_elt *ea = a->data, *eb = b->data;
 	gdouble w1, w2;
 
+	if (ea->type == ELT_LIMIT) {
+		return -1;
+	}
+	else if (eb->type == ELT_LIMIT) {
+		return 1;
+	}
+
 	/* Special logic for atoms */
 	if (ea->type == ELT_ATOM && eb->type == ELT_ATOM &&
 			ea->priority == eb->priority) {
@@ -503,8 +533,18 @@ rspamd_ast_priority_cmp (GNode *a, GNode *b)
 static gboolean
 rspamd_ast_resort_traverse (GNode *node, gpointer unused)
 {
+	GNode *children, *last;
+
 	if (node->children) {
+
+		children = node->children;
+		last = g_node_last_sibling (children);
+		/* Needed for utlist compatibility */
+		children->prev = last;
 		DL_SORT (node->children, rspamd_ast_priority_cmp);
+		/* Restore GLIB compatibility */
+		children = node->children;
+		children->prev = NULL;
 	}
 
 	return FALSE;
@@ -534,6 +574,7 @@ rspamd_parse_expression (const gchar *line, gsize len,
 	enum rspamd_expression_op op, op_stack;
 	const gchar *p, *c, *end;
 	GPtrArray *operand_stack;
+	GNode *tmp;
 
 	enum {
 		PARSE_ATOM = 0,
@@ -550,12 +591,13 @@ rspamd_parse_expression (const gchar *line, gsize len,
 	}
 
 	memset (&elt, 0, sizeof (elt));
-	num_re = rspamd_regexp_cache_create (NULL, "/^\\d+(?:\\s+|[)]|$)/", NULL, NULL);
+	num_re = rspamd_regexp_cache_create (NULL,
+			"/^(?:[+-]?([0-9]*[.])?[0-9]+)(?:\\s+|[)]|$)/", NULL, NULL);
 
 	p = line;
 	c = line;
 	end = line + len;
-	e = g_slice_alloc (sizeof (*e));
+	e = g_malloc0 (sizeof (*e));
 	e->expressions = g_array_new (FALSE, FALSE,
 			sizeof (struct rspamd_expression_elt));
 	operand_stack = g_ptr_array_sized_new (32);
@@ -571,53 +613,94 @@ rspamd_parse_expression (const gchar *line, gsize len,
 		case PARSE_ATOM:
 			if (g_ascii_isspace (*p)) {
 				state = SKIP_SPACES;
+				continue;
 			}
 			else if (rspamd_expr_is_operation_symbol (*p)) {
+				if (p + 1 < end) {
+					gchar t = *(p + 1);
+
+					if (t != ':') {
+						state = PARSE_OP;
+						continue;
+					}
+				}
+				else {
+					state = PARSE_OP;
+					continue;
+				}
+			}
+
+			/*
+			 * First of all, we check some pre-conditions:
+			 * 1) if we have 'and ' or 'or ' or 'not ' strings, they are op
+			 * 2) if we have full numeric string, then we check for
+			 * the following expression:
+			 *  ^\d+\s*[><]$
+			 */
+			if ((gulong)(end - p) > sizeof ("and ") &&
+				(g_ascii_strncasecmp (p, "and ", sizeof ("and ") - 1) == 0 ||
+				 g_ascii_strncasecmp (p, "not ", sizeof ("not ") - 1) == 0 )) {
+				state = PARSE_OP;
+			}
+			else if ((gulong)(end - p) > sizeof ("or ") &&
+					 g_ascii_strncasecmp (p, "or ", sizeof ("or ") - 1) == 0) {
 				state = PARSE_OP;
 			}
 			else {
 				/*
-				 * First of all, we check some pre-conditions:
-				 * 1) if we have 'and ' or 'or ' or 'not ' strings, they are op
-				 * 2) if we have full numeric string, then we check for
-				 * the following expression:
-				 *  ^\d+\s*[><]$
+				 * If we have any comparison operator in the stack, then try
+				 * to parse limit
 				 */
-				if ((gulong)(end - p) > sizeof ("and ") &&
-					(g_ascii_strncasecmp (p, "and ", sizeof ("and ") - 1) == 0 ||
-					g_ascii_strncasecmp (p, "not ", sizeof ("not ") - 1) == 0 )) {
-					state = PARSE_OP;
-				}
-				else if ((gulong)(end - p) > sizeof ("or ") &&
-					g_ascii_strncasecmp (p, "or ", sizeof ("or ") - 1) == 0) {
-					state = PARSE_OP;
-				}
-				else if (rspamd_regexp_search (num_re, p, end - p, NULL, NULL,
-						FALSE, NULL)) {
-					c = p;
-					state = PARSE_LIM;
-				}
-				else {
-					/* Try to parse atom */
-					atom = subr->parse (p, end - p, pool, subr_data, err);
-					if (atom == NULL) {
-						/* We couldn't parse the atom, so go out */
-						goto err;
-					}
-					g_assert (atom->len != 0);
-					p = p + atom->len;
+				op = GPOINTER_TO_INT (rspamd_expr_stack_peek (e));
 
-					/* Push to output */
-					elt.type = ELT_ATOM;
-					elt.p.atom = atom;
-					g_array_append_val (e->expressions, elt);
-					rspamd_expr_stack_elt_push (operand_stack,
-							g_node_new (rspamd_expr_dup_elt (pool, &elt)));
+				if (op >= OP_LT && op <= OP_GE) {
+					if (rspamd_regexp_search (num_re,
+							p,
+							end - p,
+							NULL,
+							NULL,
+							FALSE,
+							NULL)) {
+						c = p;
+						state = PARSE_LIM;
+						continue;
+					}
 				}
+
+				/* Try to parse atom */
+				atom = subr->parse (p, end - p, pool, subr_data, err);
+				if (atom == NULL || atom->len == 0) {
+					/* We couldn't parse the atom, so go out */
+					if (err != NULL && *err == NULL) {
+						g_set_error (err,
+								rspamd_expr_quark (),
+								500,
+								"Cannot parse atom: callback function failed"
+								" to parse '%.*s'",
+								(int) (end - p),
+								p);
+					}
+					goto err;
+				}
+
+				if (atom->str == NULL) {
+					atom->str = p;
+				}
+
+				p = p + atom->len;
+
+				/* Push to output */
+				elt.type = ELT_ATOM;
+				elt.p.atom = atom;
+				g_array_append_val (e->expressions, elt);
+				rspamd_expr_stack_elt_push (operand_stack,
+						g_node_new (rspamd_expr_dup_elt (pool, &elt)));
+
 			}
 			break;
 		case PARSE_LIM:
-			if (g_ascii_isdigit (*p) && p != end - 1) {
+			if ((g_ascii_isdigit (*p) || *p == '-' || *p == '.')
+					&& p < end - 1) {
 				p ++;
 			}
 			else {
@@ -627,7 +710,7 @@ rspamd_parse_expression (const gchar *line, gsize len,
 
 				if (p - c > 0) {
 					elt.type = ELT_LIMIT;
-					elt.p.lim.val = strtoul (c, NULL, 10);
+					elt.p.lim = strtod (c, NULL);
 					g_array_append_val (e->expressions, elt);
 					rspamd_expr_stack_elt_push (operand_stack,
 							g_node_new (rspamd_expr_dup_elt (pool, &elt)));
@@ -797,12 +880,19 @@ rspamd_parse_expression (const gchar *line, gsize len,
 	return TRUE;
 
 err:
+	while ((tmp = rspamd_expr_stack_elt_pop (operand_stack)) != NULL) {
+		g_node_destroy (tmp);
+	}
+
+	g_ptr_array_free (operand_stack, TRUE);
+	rspamd_expression_destroy (e);
+
 	return FALSE;
 }
 
 static gboolean
 rspamd_ast_node_done (struct rspamd_expression_elt *elt,
-		struct rspamd_expression_elt *parelt, gint acc, gint lim)
+		struct rspamd_expression_elt *parelt, gdouble acc, gdouble lim)
 {
 	gboolean ret = FALSE;
 
@@ -862,38 +952,39 @@ rspamd_ast_node_done (struct rspamd_expression_elt *elt,
 	return ret;
 }
 
-static gint
-rspamd_ast_do_op (struct rspamd_expression_elt *elt, gint val, gint acc, gint lim)
+static gdouble
+rspamd_ast_do_op (struct rspamd_expression_elt *elt, gdouble val,
+		gdouble acc, gdouble lim, gboolean first_elt)
 {
-	gint ret = val;
+	gdouble ret = val;
 
 	g_assert (elt->type == ELT_OP);
 
 	switch (elt->p.op) {
 	case OP_NOT:
-		ret = !val;
+		ret = fabs (val) > DOUBLE_EPSILON ? 0.0 : 1.0;
 		break;
 	case OP_PLUS:
 		ret = acc + val;
 		break;
 	case OP_GE:
-		ret = acc >= lim;
+		ret = first_elt ? (val >= lim) : (acc >= lim);
 		break;
 	case OP_GT:
-		ret = acc > lim;
+		ret = first_elt ? (val > lim) : (acc > lim);
 		break;
 	case OP_LE:
-		ret = acc <= lim;
+		ret = first_elt ? (val <= lim) : (acc <= lim);
 		break;
 	case OP_LT:
-		ret = acc < lim;
+		ret = first_elt ? (val < lim) : (acc < lim);
 		break;
 	case OP_MULT:
 	case OP_AND:
-		ret = acc && val;
+		ret = first_elt ? (val) : (acc * val);
 		break;
 	case OP_OR:
-		ret = acc || val;
+		ret = first_elt ? (val) : (acc + val);
 		break;
 	default:
 		g_assert (0);
@@ -903,14 +994,14 @@ rspamd_ast_do_op (struct rspamd_expression_elt *elt, gint val, gint acc, gint li
 	return ret;
 }
 
-static gint
-rspamd_ast_process_node (struct rspamd_expression *expr, gint flags, GNode *node,
-		gpointer data)
+static gdouble
+rspamd_ast_process_node (struct rspamd_expression *expr, GNode *node,
+						 struct rspamd_expr_process_data *process_data)
 {
 	struct rspamd_expression_elt *elt, *celt, *parelt = NULL;
 	GNode *cld;
-	gint acc = G_MININT, lim = G_MININT, val;
-	gdouble t1, t2;
+	gdouble acc = NAN, lim = 0;
+	gdouble t1, t2, val;
 	gboolean calc_ticks = FALSE;
 
 	elt = node->data;
@@ -927,17 +1018,21 @@ rspamd_ast_process_node (struct rspamd_expression *expr, gint flags, GNode *node
 			 */
 			if ((expr->evals & 0x1F) == (GPOINTER_TO_UINT (node) >> 4 & 0x1F)) {
 				calc_ticks = TRUE;
-				t1 = rspamd_get_ticks ();
+				t1 = rspamd_get_ticks (TRUE);
 			}
 
-			elt->value = expr->subr->process (data, elt->p.atom);
+			elt->value = expr->subr->process (process_data, elt->p.atom);
 
-			if (elt->value) {
+			if (fabs (elt->value) > 1e-9) {
 				elt->p.atom->hits ++;
+
+				if (process_data->trace) {
+					g_ptr_array_add (process_data->trace, elt->p.atom);
+				}
 			}
 
 			if (calc_ticks) {
-				t2 = rspamd_get_ticks ();
+				t2 = rspamd_get_ticks (TRUE);
 				elt->p.atom->avg_ticks += ((t2 - t1) - elt->p.atom->avg_ticks) /
 						(expr->evals);
 			}
@@ -945,15 +1040,14 @@ rspamd_ast_process_node (struct rspamd_expression *expr, gint flags, GNode *node
 			elt->flags |= RSPAMD_EXPR_FLAG_PROCESSED;
 		}
 
-		return elt->value;
+		acc = elt->value;
 		break;
 	case ELT_LIMIT:
 
-		return elt->p.lim.val;
+		acc = elt->p.lim;
 		break;
 	case ELT_OP:
 		g_assert (node->children != NULL);
-		cld = node->children;
 
 		/* Try to find limit at the parent node */
 		if (node->parent) {
@@ -961,7 +1055,7 @@ rspamd_ast_process_node (struct rspamd_expression *expr, gint flags, GNode *node
 			celt = node->parent->children->data;
 
 			if (celt->type == ELT_LIMIT) {
-				lim = celt->p.lim.val;
+				lim = celt->p.lim;
 			}
 		}
 
@@ -970,19 +1064,20 @@ rspamd_ast_process_node (struct rspamd_expression *expr, gint flags, GNode *node
 
 			/* Save limit if we've found it */
 			if (celt->type == ELT_LIMIT) {
-				lim = celt->p.lim.val;
+				lim = celt->p.lim;
 				continue;
 			}
 
-			val = rspamd_ast_process_node (expr, flags, cld, data);
+			val = rspamd_ast_process_node (expr, cld, process_data);
 
-			if (acc == G_MININT) {
-				acc = val;
+			if (isnan (acc)) {
+				acc = rspamd_ast_do_op (elt, val, 0, lim, TRUE);
+			}
+			else {
+				acc = rspamd_ast_do_op (elt, val, acc, lim, FALSE);
 			}
 
-			acc = rspamd_ast_do_op (elt, val, acc, lim);
-
-			if (!(flags & RSPAMD_EXPRESSION_FLAG_NOOPT)) {
+			if (!(process_data->flags & RSPAMD_EXPRESSION_FLAG_NOOPT)) {
 				if (rspamd_ast_node_done (elt, parelt, acc, lim)) {
 					return acc;
 				}
@@ -1005,26 +1100,24 @@ rspamd_ast_cleanup_traverse (GNode *n, gpointer d)
 	return FALSE;
 }
 
-gint
-rspamd_process_expression (struct rspamd_expression *expr, gint flags,
-		gpointer data)
+gdouble
+rspamd_process_expression_track (struct rspamd_expression *expr, struct rspamd_expr_process_data *process_data)
 {
-	gint ret = 0;
+	gdouble ret = 0;
 
 	g_assert (expr != NULL);
 	/* Ensure that stack is empty at this point */
 	g_assert (expr->expression_stack->len == 0);
 
-	ret = rspamd_ast_process_node (expr, flags, expr->ast, data);
+	expr->evals ++;
+	ret = rspamd_ast_process_node (expr, expr->ast, process_data);
 
 	/* Cleanup */
 	g_node_traverse (expr->ast, G_IN_ORDER, G_TRAVERSE_ALL, -1,
 			rspamd_ast_cleanup_traverse, NULL);
 
-	expr->evals ++;
-
 	/* Check if we need to resort */
-	if (expr->evals == expr->next_resort) {
+	if (expr->evals % expr->next_resort == 0) {
 		expr->next_resort = ottery_rand_range (MAX_RESORT_EVALS) +
 				MIN_RESORT_EVALS;
 		/* Set priorities for branches */
@@ -1039,6 +1132,12 @@ rspamd_process_expression (struct rspamd_expression *expr, gint flags,
 	return ret;
 }
 
+gdouble
+rspamd_process_expression (struct rspamd_expression *expr, struct rspamd_expr_process_data *process_data)
+{
+	return rspamd_process_expression_track (expr, process_data);
+}
+
 static gboolean
 rspamd_ast_string_traverse (GNode *n, gpointer d)
 {
@@ -1049,10 +1148,16 @@ rspamd_ast_string_traverse (GNode *n, gpointer d)
 	const char *op_str = NULL;
 
 	if (elt->type == ELT_ATOM) {
-		g_string_append_len (res, elt->p.atom->str, elt->p.atom->len);
+		rspamd_printf_gstring (res, "(%*s)",
+				(int)elt->p.atom->len, elt->p.atom->str);
 	}
 	else if (elt->type == ELT_LIMIT) {
-		rspamd_printf_gstring (res, "%d", elt->p.lim.val);
+		if (elt->p.lim == (double)(gint64)elt->p.lim) {
+			rspamd_printf_gstring (res, "%L", (gint64)elt->p.lim);
+		}
+		else {
+			rspamd_printf_gstring (res, "%f", elt->p.lim);
+		}
 	}
 	else {
 		op_str = rspamd_expr_op_to_str (elt->p.op);
@@ -1091,3 +1196,55 @@ rspamd_expression_tostring (struct rspamd_expression *expr)
 
 	return res;
 }
+
+struct atom_foreach_cbdata {
+	rspamd_expression_atom_foreach_cb cb;
+	gpointer cbdata;
+};
+
+static gboolean
+rspamd_ast_atom_traverse (GNode *n, gpointer d)
+{
+	struct atom_foreach_cbdata *data = d;
+	struct rspamd_expression_elt *elt = n->data;
+	rspamd_ftok_t tok;
+
+	if (elt->type == ELT_ATOM) {
+		tok.begin = elt->p.atom->str;
+		tok.len = elt->p.atom->len;
+
+		data->cb (&tok, data->cbdata);
+	}
+
+	return FALSE;
+}
+
+void
+rspamd_expression_atom_foreach (struct rspamd_expression *expr,
+		rspamd_expression_atom_foreach_cb cb, gpointer cbdata)
+{
+	struct atom_foreach_cbdata data;
+
+	g_assert (expr != NULL);
+
+	data.cb = cb;
+	data.cbdata = cbdata;
+	g_node_traverse (expr->ast, G_POST_ORDER, G_TRAVERSE_ALL, -1,
+			rspamd_ast_atom_traverse, &data);
+}
+
+gboolean
+rspamd_expression_node_is_op (GNode *node, enum rspamd_expression_op op)
+{
+	struct rspamd_expression_elt *elt;
+
+	g_assert (node != NULL);
+
+	elt = node->data;
+
+	if (elt->type == ELT_OP && elt->p.op == op) {
+		return TRUE;
+	}
+
+	return FALSE;
+}
