@@ -1,31 +1,23 @@
-/* Copyright (c) 2010, Vsevolod Stakhov
- * All rights reserved.
+/*-
+ * Copyright 2016 Vsevolod Stakhov
  *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions are met:
- *       * Redistributions of source code must retain the above copyright
- *         notice, this list of conditions and the following disclaimer.
- *       * Redistributions in binary form must reproduce the above copyright
- *         notice, this list of conditions and the following disclaimer in the
- *         documentation and/or other materials provided with the distribution.
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
  *
- * THIS SOFTWARE IS PROVIDED ''AS IS'' AND ANY
- * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
- * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
- * DISCLAIMED. IN NO EVENT SHALL Rambler BE LIABLE FOR ANY
- * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
- * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
- * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
- * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
- * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
- * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ *   http://www.apache.org/licenses/LICENSE-2.0
+ *
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
  */
-
-
 #ifndef PRINTF_H_
 #define PRINTF_H_
 
 #include "config.h"
+#include "fstring.h"
 
 /*
  * supported formats:
@@ -40,18 +32,21 @@
  *	%[0][width][.width]F	    long double
  *	%[0][width][.width]g	    double
  *	%[0][width][.width]G	    long double
- *	%b                          boolean (true or false)
  *	%P						    pid_t
  *	%r				            rlim_t
  *	%p						    void *
- *	%V						    f_str_t *
+ *	%V						    rspamd_fstring_t *
+ *	%T						    rspamd_ftok_t
  *	%v                          GString *
  *	%s						    null-terminated string
  *	%xs                         hex encoded string
+ *	%bs                         base32 encoded string
+ *	%Bs							base64 encoded string
  *	%*s					        length and string
  *	%Z						    '\0'
  *	%N						    '\n'
  *	%c						    gchar
+ *	%t						    time_t
  *	%e                          GError *
  *	%%						    %
  *
@@ -60,7 +55,7 @@
 /**
  * Callback used for common printf operations
  * @param buf buffer to append
- * @param buflen lenght of the buffer
+ * @param buflen length of the buffer
  * @param ud opaque pointer
  * @return number of characters written
  */
@@ -76,6 +71,9 @@ gchar * rspamd_vsnprintf (gchar *buf, glong max, const gchar *fmt,
 glong rspamd_printf_gstring (GString *s, const gchar *fmt, ...);
 glong rspamd_vprintf_gstring (GString *s, const gchar *fmt, va_list args);
 
+glong rspamd_printf_fstring (rspamd_fstring_t **s, const gchar *fmt, ...);
+glong rspamd_vprintf_fstring (rspamd_fstring_t **s, const gchar *fmt, va_list args);
+
 glong rspamd_vprintf_common (rspamd_printf_append_func func,
 	gpointer apd,
 	const gchar *fmt,
