@@ -1,26 +1,18 @@
-/* Copyright (c) 2013, Vsevolod Stakhov
- * All rights reserved.
+/*-
+ * Copyright 2016 Vsevolod Stakhov
  *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions are met:
- *       * Redistributions of source code must retain the above copyright
- *         notice, this list of conditions and the following disclaimer.
- *       * Redistributions in binary form must reproduce the above copyright
- *         notice, this list of conditions and the following disclaimer in the
- *         documentation and/or other materials provided with the distribution.
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
  *
- * THIS SOFTWARE IS PROVIDED ''AS IS'' AND ANY
- * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
- * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
- * DISCLAIMED. IN NO EVENT SHALL AUTHOR BE LIABLE FOR ANY
- * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
- * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
- * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
- * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
- * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
- * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ *   http://www.apache.org/licenses/LICENSE-2.0
+ *
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
  */
-
 #include "lua_common.h"
 
 /***
@@ -59,8 +51,9 @@ print_octets(ip6)
  */
 
 /***
- * @method ip:to_string()
+ * @method ip:to_string([pretty=false])
  * Converts valid IP address to string
+ * @param {bool} pretty print IP address with port and braces (for IPv6)
  * @return {string or nil} string representation of IP or `nil` if IP is invalid
  */
 LUA_FUNCTION_DEF (ip, to_string);
@@ -139,7 +132,7 @@ LUA_FUNCTION_DEF (ip, get_version);
 LUA_FUNCTION_DEF (ip, is_valid);
 /***
  * @method ip:apply_mask(mask)
- * Applies mask to IP address, reseting up to `mask` least significant bits to zero.
+ * Applies mask to IP address, resetting up to `mask` least significant bits to zero.
  * @param {integer} mask how many bits to reset
  * @return {ip} new IP object with `mask` bits reset
  */
@@ -164,6 +157,12 @@ LUA_FUNCTION_DEF (ip, copy);
  * @return {number} port number or nil
  */
 LUA_FUNCTION_DEF (ip, get_port);
+/***
+ * @method ip:is_local()
+ * Returns true if address is local one
+ * @return {boolean} `true` if address is local
+ */
+LUA_FUNCTION_DEF (ip, is_local);
 
 static const struct luaL_reg iplib_m[] = {
 	LUA_INTERFACE_DEF (ip, to_string),
@@ -176,6 +175,10 @@ static const struct luaL_reg iplib_m[] = {
 	LUA_INTERFACE_DEF (ip, is_valid),
 	LUA_INTERFACE_DEF (ip, apply_mask),
 	LUA_INTERFACE_DEF (ip, copy),
+	LUA_INTERFACE_DEF (ip, is_local),
+	{"tostring", lua_ip_to_string},
+	{"totable", lua_ip_to_table},
+	{"tonumber", lua_ip_to_number},
 	{"__tostring", lua_ip_to_string},
 	{"__eq", lua_ip_equal},
 	{"__gc", lua_ip_destroy},
@@ -193,7 +196,7 @@ lua_ip_new (lua_State *L, struct rspamd_lua_ip *old)
 {
 	struct rspamd_lua_ip *ip, **pip;
 
-	ip = g_slice_alloc (sizeof (*ip));
+	ip = g_malloc0 (sizeof (*ip));
 
 	if (old != NULL && old->addr != NULL) {
 		ip->addr = rspamd_inet_address_copy (old->addr);
@@ -210,7 +213,7 @@ lua_ip_new (lua_State *L, struct rspamd_lua_ip *old)
 struct rspamd_lua_ip *
 lua_check_ip (lua_State * L, gint pos)
 {
-	void *ud = luaL_checkudata (L, pos, "rspamd{ip}");
+	void *ud = rspamd_lua_check_udata (L, pos, "rspamd{ip}");
 
 	luaL_argcheck (L, ud != NULL, pos, "'ip' expected");
 	return ud ? *((struct rspamd_lua_ip **)ud) : NULL;
@@ -219,16 +222,17 @@ lua_check_ip (lua_State * L, gint pos)
 static gint
 lua_ip_to_table (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct rspamd_lua_ip *ip = lua_check_ip (L, 1);
 	guint max, i;
 	guint8 *ptr;
 
 	if (ip != NULL && ip->addr) {
-		lua_newtable (L);
-		ptr = rspamd_inet_address_get_radix_key (ip->addr, &max);
+		ptr = rspamd_inet_address_get_hash_key (ip->addr, &max);
+		lua_createtable (L, max, 0);
 
 		for (i = 1; i <= max; i++, ptr++) {
-			lua_pushnumber (L, *ptr);
+			lua_pushinteger (L, *ptr);
 			lua_rawseti (L, -2, i);
 		}
 	}
@@ -242,6 +246,7 @@ lua_ip_to_table (lua_State *L)
 static gint
 lua_ip_str_octets (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct rspamd_lua_ip *ip = lua_check_ip (L, 1);
 	guint max, i;
 	guint8 *ptr;
@@ -249,9 +254,9 @@ lua_ip_str_octets (lua_State *L)
 	char numbuf[8];
 
 	if (ip != NULL && ip->addr) {
-		lua_newtable (L);
 		af = rspamd_inet_address_get_af (ip->addr);
-		ptr = rspamd_inet_address_get_radix_key (ip->addr, &max);
+		ptr = rspamd_inet_address_get_hash_key (ip->addr, &max);
+		lua_createtable (L, max * 2, 0);
 
 		for (i = 1; i <= max; i++, ptr++) {
 			if (af == AF_INET) {
@@ -282,6 +287,7 @@ lua_ip_str_octets (lua_State *L)
 static gint
 lua_ip_inversed_str_octets (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct rspamd_lua_ip *ip = lua_check_ip (L, 1);
 	guint max, i;
 	guint8 *ptr;
@@ -289,9 +295,9 @@ lua_ip_inversed_str_octets (lua_State *L)
 	gint af;
 
 	if (ip != NULL && ip->addr) {
-		lua_newtable (L);
-		ptr = rspamd_inet_address_get_radix_key (ip->addr, &max);
+		ptr = rspamd_inet_address_get_hash_key (ip->addr, &max);
 		af = rspamd_inet_address_get_af (ip->addr);
+		lua_createtable (L, max * 2, 0);
 
 		ptr += max - 1;
 		for (i = 1; i <= max; i++, ptr--) {
@@ -323,13 +329,19 @@ lua_ip_inversed_str_octets (lua_State *L)
 static gint
 lua_ip_to_string (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct rspamd_lua_ip *ip = lua_check_ip (L, 1);
 
 	if (ip != NULL && ip->addr) {
-		lua_pushstring (L, rspamd_inet_address_to_string (ip->addr));
+		if (lua_isboolean (L, 2) && lua_toboolean (L, 2) == true) {
+			lua_pushstring (L, rspamd_inet_address_to_string_pretty (ip->addr));
+		}
+		else {
+			lua_pushstring (L, rspamd_inet_address_to_string (ip->addr));
+		}
 	}
 	else {
-		lua_pushnil (L);
+		luaL_error (L, "invalid arguments");
 	}
 
 	return 1;
@@ -338,10 +350,11 @@ lua_ip_to_string (lua_State *L)
 static gint
 lua_ip_get_port (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct rspamd_lua_ip *ip = lua_check_ip (L, 1);
 
 	if (ip != NULL && ip->addr) {
-		lua_pushnumber (L, rspamd_inet_address_get_port (ip->addr));
+		lua_pushinteger (L, rspamd_inet_address_get_port (ip->addr));
 	}
 	else {
 		lua_pushnil (L);
@@ -353,13 +366,18 @@ lua_ip_get_port (lua_State *L)
 static gint
 lua_ip_from_string (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct rspamd_lua_ip *ip;
 	const gchar *ip_str;
 
 	ip_str = luaL_checkstring (L, 1);
 	if (ip_str) {
 		ip = lua_ip_new (L, NULL);
-		rspamd_parse_inet_address (&ip->addr, ip_str);
+
+		if (!rspamd_parse_inet_address (&ip->addr, ip_str, 0)) {
+			msg_warn ("cannot parse ip: %s", ip_str);
+			ip->addr = NULL;
+		}
 	}
 	else {
 		lua_pushnil (L);
@@ -371,13 +389,14 @@ lua_ip_from_string (lua_State *L)
 static gint
 lua_ip_to_number (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct rspamd_lua_ip *ip = lua_check_ip (L, 1);
 	guint32 c;
 	guint max, i;
 	guchar *ptr;
 
 	if (ip != NULL && ip->addr) {
-		ptr = rspamd_inet_address_get_radix_key (ip->addr, &max);
+		ptr = rspamd_inet_address_get_hash_key (ip->addr, &max);
 
 		for (i = 0; i < max / sizeof (c); i ++) {
 			memcpy (&c, ptr + i * sizeof (c), sizeof (c));
@@ -397,13 +416,14 @@ lua_ip_to_number (lua_State *L)
 static gint
 lua_ip_destroy (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct rspamd_lua_ip *ip = lua_check_ip (L, 1);
 
 	if (ip) {
 		if (ip->addr) {
-			rspamd_inet_address_destroy (ip->addr);
+			rspamd_inet_address_free (ip->addr);
 		}
-		g_slice_free1 (sizeof (struct rspamd_lua_ip), ip);
+		g_free (ip);
 	}
 
 	return 0;
@@ -412,10 +432,11 @@ lua_ip_destroy (lua_State *L)
 static gint
 lua_ip_get_version (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct rspamd_lua_ip *ip = lua_check_ip (L, 1);
 
 	if (ip && ip->addr) {
-		lua_pushnumber (L, rspamd_inet_address_get_af (ip->addr) == AF_INET6 ?
+		lua_pushinteger (L, rspamd_inet_address_get_af (ip->addr) == AF_INET6 ?
 				6 : 4);
 	}
 	else {
@@ -428,6 +449,7 @@ lua_ip_get_version (lua_State *L)
 static gint
 lua_ip_is_valid (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct rspamd_lua_ip *ip = lua_check_ip (L, 1);
 
 	if (ip) {
@@ -443,6 +465,7 @@ lua_ip_is_valid (lua_State *L)
 static gint
 lua_ip_apply_mask (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct rspamd_lua_ip *ip = lua_check_ip (L, 1), *nip;
 	gint mask;
 
@@ -461,12 +484,13 @@ lua_ip_apply_mask (lua_State *L)
 static gint
 lua_ip_equal (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct rspamd_lua_ip *ip1 = lua_check_ip (L, 1),
 		*ip2 = lua_check_ip (L, 2);
 	gboolean res = FALSE;
 
 	if (ip1 && ip2 && ip1->addr && ip2->addr) {
-		res = rspamd_inet_address_compare (ip1->addr, ip2->addr);
+		res = rspamd_inet_address_compare (ip1->addr, ip2->addr, TRUE);
 	}
 
 	lua_pushboolean (L, res);
@@ -477,6 +501,7 @@ lua_ip_equal (lua_State *L)
 static gint
 lua_ip_copy (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct rspamd_lua_ip *ip = lua_check_ip (L, 1);
 
 	if (ip) {
@@ -489,12 +514,34 @@ lua_ip_copy (lua_State *L)
 	return 1;
 }
 
+static gint
+lua_ip_is_local (lua_State *L)
+{
+	struct rspamd_lua_ip *ip = lua_check_ip (L, 1);
+	gboolean check_laddrs = TRUE;
+
+	if (ip && ip->addr) {
+
+		if (lua_type (L, 2) == LUA_TBOOLEAN) {
+			check_laddrs = lua_toboolean (L, 2);
+		}
+
+		lua_pushboolean (L, rspamd_inet_address_is_local (ip->addr,
+				check_laddrs));
+	}
+	else {
+		lua_pushnil (L);
+	}
+
+	return 1;
+}
+
 void
 rspamd_lua_ip_push (lua_State *L, rspamd_inet_addr_t *addr)
 {
 	struct rspamd_lua_ip *ip, **pip;
 
-	ip = g_slice_alloc0 (sizeof (struct rspamd_lua_ip));
+	ip = g_malloc0 (sizeof (struct rspamd_lua_ip));
 	ip->addr = rspamd_inet_address_copy (addr);
 	pip = lua_newuserdata (L, sizeof (struct rspamd_lua_ip *));
 	rspamd_lua_setclass (L, "rspamd{ip}", -1);
@@ -510,12 +557,18 @@ rspamd_lua_ip_push_fromstring (lua_State *L, const gchar *ip_str)
 		lua_pushnil (L);
 	}
 	else {
-		ip = g_slice_alloc0 (sizeof (struct rspamd_lua_ip));
-		rspamd_parse_inet_address (&ip->addr, ip_str);
+		ip = g_malloc0 (sizeof (struct rspamd_lua_ip));
 
-		pip = lua_newuserdata (L, sizeof (struct rspamd_lua_ip *));
-		rspamd_lua_setclass (L, "rspamd{ip}", -1);
-		*pip = ip;
+		if (rspamd_parse_inet_address (&ip->addr, ip_str, 0)) {
+
+			pip = lua_newuserdata (L, sizeof (struct rspamd_lua_ip *));
+			rspamd_lua_setclass (L, "rspamd{ip}", -1);
+			*pip = ip;
+		}
+		else {
+			g_free (ip);
+			lua_pushnil (L);
+		}
 	}
 }
 
@@ -543,5 +596,5 @@ luaopen_ip (lua_State * L)
 	luaL_register (L, NULL,		   iplib_m);
 	rspamd_lua_add_preload (L, "rspamd_ip", lua_load_ip);
 
-	lua_pop (L, 1);                      /* remove metatable from stack */
+	lua_pop (L, 1);
 }
