@@ -1,28 +1,19 @@
-/* Copyright (c) 2010, Vsevolod Stakhov
- * All rights reserved.
+/*-
+ * Copyright 2016 Vsevolod Stakhov
  *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions are met:
- *       * Redistributions of source code must retain the above copyright
- *         notice, this list of conditions and the following disclaimer.
- *       * Redistributions in binary form must reproduce the above copyright
- *         notice, this list of conditions and the following disclaimer in the
- *         documentation and/or other materials provided with the distribution.
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
  *
- * THIS SOFTWARE IS PROVIDED ''AS IS'' AND ANY
- * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
- * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
- * DISCLAIMED. IN NO EVENT SHALL AUTHOR BE LIABLE FOR ANY
- * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
- * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
- * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
- * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
- * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
- * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ *   http://www.apache.org/licenses/LICENSE-2.0
+ *
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
  */
-
 #include "lua_common.h"
-#include "regexp.h"
 
 /***
  * @module rspamd_regexp
@@ -38,10 +29,14 @@
  */
 
 LUA_FUNCTION_DEF (regexp, create);
+LUA_FUNCTION_DEF (regexp, import_glob);
+LUA_FUNCTION_DEF (regexp, import_plain);
 LUA_FUNCTION_DEF (regexp, create_cached);
 LUA_FUNCTION_DEF (regexp, get_cached);
 LUA_FUNCTION_DEF (regexp, get_pattern);
 LUA_FUNCTION_DEF (regexp, set_limit);
+LUA_FUNCTION_DEF (regexp, set_max_hits);
+LUA_FUNCTION_DEF (regexp, get_max_hits);
 LUA_FUNCTION_DEF (regexp, search);
 LUA_FUNCTION_DEF (regexp, match);
 LUA_FUNCTION_DEF (regexp, matchn);
@@ -52,6 +47,8 @@ LUA_FUNCTION_DEF (regexp, gc);
 static const struct luaL_reg regexplib_m[] = {
 	LUA_INTERFACE_DEF (regexp, get_pattern),
 	LUA_INTERFACE_DEF (regexp, set_limit),
+	LUA_INTERFACE_DEF (regexp, set_max_hits),
+	LUA_INTERFACE_DEF (regexp, get_max_hits),
 	LUA_INTERFACE_DEF (regexp, match),
 	LUA_INTERFACE_DEF (regexp, matchn),
 	LUA_INTERFACE_DEF (regexp, search),
@@ -63,6 +60,8 @@ static const struct luaL_reg regexplib_m[] = {
 };
 static const struct luaL_reg regexplib_f[] = {
 	LUA_INTERFACE_DEF (regexp, create),
+	LUA_INTERFACE_DEF (regexp, import_glob),
+	LUA_INTERFACE_DEF (regexp, import_plain),
 	LUA_INTERFACE_DEF (regexp, get_cached),
 	LUA_INTERFACE_DEF (regexp, create_cached),
 	{NULL, NULL}
@@ -73,22 +72,46 @@ static const struct luaL_reg regexplib_f[] = {
 
 rspamd_mempool_t *regexp_static_pool = NULL;
 
-struct rspamd_lua_regexp {
-	rspamd_regexp_t *re;
-	gchar *re_pattern;
-	gsize match_limit;
-	gint re_flags;
-};
-
 static struct rspamd_lua_regexp *
 lua_check_regexp (lua_State * L)
 {
-	void *ud = luaL_checkudata (L, 1, "rspamd{regexp}");
+	void *ud = rspamd_lua_check_udata (L, 1, "rspamd{regexp}");
 
 	luaL_argcheck (L, ud != NULL, 1, "'regexp' expected");
 	return ud ? *((struct rspamd_lua_regexp **)ud) : NULL;
 }
 
+static gchar *
+rspamd_lua_get_module_name (lua_State *L)
+{
+	lua_Debug d;
+	gchar *p;
+	gchar func_buf[128];
+
+	if (lua_getstack (L, 1, &d) == 1) {
+		(void) lua_getinfo (L, "Sl", &d);
+		if ((p = strrchr (d.short_src, '/')) == NULL) {
+			p = d.short_src;
+		}
+		else {
+			p++;
+		}
+
+		if (strlen (p) > 20) {
+			rspamd_snprintf (func_buf, sizeof (func_buf), "%10s...]:%d", p,
+					d.currentline);
+		}
+		else {
+			rspamd_snprintf (func_buf, sizeof (func_buf), "%s:%d", p,
+					d.currentline);
+		}
+
+		return g_strdup (func_buf);
+	}
+
+	return NULL;
+}
+
 /***
  * @function rspamd_regexp.create(pattern[, flags])
  * Creates new rspamd_regexp
@@ -103,6 +126,7 @@ lua_check_regexp (lua_State * L)
 static int
 lua_regexp_create (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	rspamd_regexp_t *re;
 	struct rspamd_lua_regexp *new, **pnew;
 	const gchar *string, *flags_str = NULL;
@@ -113,20 +137,144 @@ lua_regexp_create (lua_State *L)
 		flags_str = luaL_checkstring (L, 2);
 	}
 
-	re = rspamd_regexp_new (string, flags_str, &err);
-	if (re == NULL) {
-		lua_pushnil (L);
-		msg_info ("cannot parse regexp: %s, error: %s",
-			string,
-			err == NULL ? "undefined" : err->message);
-		g_error_free (err);
+	if (string) {
+		re = rspamd_regexp_new (string, flags_str, &err);
+		if (re == NULL) {
+			lua_pushnil (L);
+			msg_info ("cannot parse regexp: %s, error: %s",
+					string,
+					err == NULL ? "undefined" : err->message);
+			g_error_free (err);
+		} else {
+			new = g_malloc0 (sizeof (struct rspamd_lua_regexp));
+			new->re = re;
+			new->re_pattern = g_strdup (string);
+			new->module = rspamd_lua_get_module_name (L);
+			pnew = lua_newuserdata (L, sizeof (struct rspamd_lua_regexp *));
+			rspamd_lua_setclass (L, "rspamd{regexp}", -1);
+			*pnew = new;
+		}
 	}
 	else {
-		new = g_slice_alloc0 (sizeof (struct rspamd_lua_regexp));
-		new->re = re;
-		pnew = lua_newuserdata (L, sizeof (struct rspamd_lua_regexp *));
-		rspamd_lua_setclass (L, "rspamd{regexp}", -1);
-		*pnew = new;
+		return luaL_error (L, "invalid arguments");
+	}
+
+	return 1;
+}
+
+/***
+ * @function rspamd_regexp.import_glob(glob_pattern[, flags])
+ * Creates new rspamd_regexp from glob
+ * @param {string} pattern pattern to build regexp.
+ * @param {string} flags optional flags to create regular expression
+ * @return {regexp} regexp argument that is *not* automatically destroyed
+ * @example
+ * local regexp = require "rspamd_regexp"
+ *
+ * local re = regexp.import_glob('ab*', 'i')
+ */
+static int
+lua_regexp_import_glob (lua_State *L)
+{
+	LUA_TRACE_POINT;
+	rspamd_regexp_t *re;
+	struct rspamd_lua_regexp *new, **pnew;
+	const gchar *string, *flags_str = NULL;
+	gchar *escaped;
+	gsize pat_len;
+	GError *err = NULL;
+
+	string = luaL_checklstring (L, 1, &pat_len);
+
+	if (lua_gettop (L) == 2) {
+		flags_str = luaL_checkstring (L, 2);
+	}
+
+	if (string) {
+		escaped = rspamd_str_regexp_escape (string, pat_len, NULL,
+				RSPAMD_REGEXP_ESCAPE_GLOB|RSPAMD_REGEXP_ESCAPE_UTF);
+
+		re = rspamd_regexp_new (escaped, flags_str, &err);
+
+		if (re == NULL) {
+			lua_pushnil (L);
+			msg_info ("cannot parse regexp: %s, error: %s",
+					string,
+					err == NULL ? "undefined" : err->message);
+			g_error_free (err);
+			g_free (escaped);
+		}
+		else {
+			new = g_malloc0 (sizeof (struct rspamd_lua_regexp));
+			new->re = re;
+			new->re_pattern = escaped;
+			new->module = rspamd_lua_get_module_name (L);
+			pnew = lua_newuserdata (L, sizeof (struct rspamd_lua_regexp *));
+			rspamd_lua_setclass (L, "rspamd{regexp}", -1);
+			*pnew = new;
+		}
+	}
+	else {
+		return luaL_error (L, "invalid arguments");
+	}
+
+	return 1;
+}
+
+/***
+ * @function rspamd_regexp.import_plain(plain_string[, flags])
+ * Creates new rspamd_regexp from plain string (escaping specials)
+ * @param {string} pattern pattern to build regexp.
+ * @param {string} flags optional flags to create regular expression
+ * @return {regexp} regexp argument that is *not* automatically destroyed
+ * @example
+ * local regexp = require "rspamd_regexp"
+ *
+ * local re = regexp.import_plain('exact_string_with*', 'i')
+ */
+static int
+lua_regexp_import_plain (lua_State *L)
+{
+	LUA_TRACE_POINT;
+	rspamd_regexp_t *re;
+	struct rspamd_lua_regexp *new, **pnew;
+	const gchar *string, *flags_str = NULL;
+	gchar *escaped;
+	gsize pat_len;
+	GError *err = NULL;
+
+	string = luaL_checklstring (L, 1, &pat_len);
+
+	if (lua_gettop (L) == 2) {
+		flags_str = luaL_checkstring (L, 2);
+	}
+
+	if (string) {
+		escaped = rspamd_str_regexp_escape (string, pat_len, NULL,
+				RSPAMD_REGEXP_ESCAPE_ASCII);
+
+		re = rspamd_regexp_new (escaped, flags_str, &err);
+
+		if (re == NULL) {
+			lua_pushnil (L);
+			msg_info ("cannot parse regexp: %s, error: %s",
+					string,
+					err == NULL ? "undefined" : err->message);
+			g_error_free (err);
+			g_free (escaped);
+		}
+		else {
+			new = g_malloc0 (sizeof (struct rspamd_lua_regexp));
+			new->re = re;
+			new->re_pattern = escaped;
+			new->module = rspamd_lua_get_module_name (L);
+			pnew = lua_newuserdata (L, sizeof (struct rspamd_lua_regexp *));
+			rspamd_lua_setclass (L, "rspamd{regexp}", -1);
+			*pnew = new;
+		}
+	}
+	else {
+		return luaL_error (L, "invalid arguments");
 	}
 
 	return 1;
@@ -143,6 +291,7 @@ lua_regexp_create (lua_State *L)
 static int
 lua_regexp_get_cached (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	rspamd_regexp_t *re;
 	struct rspamd_lua_regexp *new, **pnew;
 	const gchar *string, *flags_str = NULL;
@@ -152,17 +301,24 @@ lua_regexp_get_cached (lua_State *L)
 		flags_str = luaL_checkstring (L, 2);
 	}
 
-	re = rspamd_regexp_cache_query (NULL, string, flags_str);
+	if (string) {
+		re = rspamd_regexp_cache_query (NULL, string, flags_str);
 
-	if (re) {
-		new = g_slice_alloc0 (sizeof (struct rspamd_lua_regexp));
-		new->re = rspamd_regexp_ref (re);
-		pnew = lua_newuserdata (L, sizeof (struct rspamd_lua_regexp *));
-		rspamd_lua_setclass (L, "rspamd{regexp}", -1);
-		*pnew = new;
+		if (re) {
+			new = g_malloc0 (sizeof (struct rspamd_lua_regexp));
+			new->re = rspamd_regexp_ref (re);
+			new->re_pattern = g_strdup (string);
+			new->module = rspamd_lua_get_module_name (L);
+			pnew = lua_newuserdata (L, sizeof (struct rspamd_lua_regexp *));
+			rspamd_lua_setclass (L, "rspamd{regexp}", -1);
+			*pnew = new;
+		}
+		else {
+			lua_pushnil (L);
+		}
 	}
 	else {
-		lua_pushnil (L);
+		return luaL_error (L, "invalid arguments");
 	}
 
 	return 1;
@@ -186,6 +342,7 @@ lua_regexp_get_cached (lua_State *L)
 static int
 lua_regexp_create_cached (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	rspamd_regexp_t *re;
 	struct rspamd_lua_regexp *new, **pnew;
 	const gchar *string, *flags_str = NULL;
@@ -196,32 +353,40 @@ lua_regexp_create_cached (lua_State *L)
 		flags_str = luaL_checkstring (L, 2);
 	}
 
-	re = rspamd_regexp_cache_query (NULL, string, flags_str);
+	if (string) {
+		re = rspamd_regexp_cache_query (NULL, string, flags_str);
 
-	if (re) {
-		new = g_slice_alloc0 (sizeof (struct rspamd_lua_regexp));
-		new->re = rspamd_regexp_ref (re);
-		pnew = lua_newuserdata (L, sizeof (struct rspamd_lua_regexp *));
-
-		rspamd_lua_setclass (L, "rspamd{regexp}", -1);
-		*pnew = new;
-	}
-	else {
-		re = rspamd_regexp_cache_create (NULL, string, flags_str, &err);
-		if (re == NULL) {
-			lua_pushnil (L);
-			msg_info ("cannot parse regexp: %s, error: %s",
-					string,
-					err == NULL ? "undefined" : err->message);
-			g_error_free (err);
-		}
-		else {
-			new = g_slice_alloc0 (sizeof (struct rspamd_lua_regexp));
+		if (re) {
+			new = g_malloc0 (sizeof (struct rspamd_lua_regexp));
 			new->re = rspamd_regexp_ref (re);
+			new->re_pattern = g_strdup (string);
+			new->module = rspamd_lua_get_module_name (L);
 			pnew = lua_newuserdata (L, sizeof (struct rspamd_lua_regexp *));
+
 			rspamd_lua_setclass (L, "rspamd{regexp}", -1);
 			*pnew = new;
 		}
+		else {
+			re = rspamd_regexp_cache_create (NULL, string, flags_str, &err);
+			if (re == NULL) {
+				lua_pushnil (L);
+				msg_info ("cannot parse regexp: %s, error: %s",
+						string,
+						err == NULL ? "undefined" : err->message);
+				g_error_free (err);
+			} else {
+				new = g_malloc0 (sizeof (struct rspamd_lua_regexp));
+				new->re = rspamd_regexp_ref (re);
+				new->re_pattern = g_strdup (string);
+				new->module = rspamd_lua_get_module_name (L);
+				pnew = lua_newuserdata (L, sizeof (struct rspamd_lua_regexp *));
+				rspamd_lua_setclass (L, "rspamd{regexp}", -1);
+				*pnew = new;
+			}
+		}
+	}
+	else {
+		return luaL_error (L, "invalid arguments");
 	}
 
 	return 1;
@@ -235,6 +400,7 @@ lua_regexp_create_cached (lua_State *L)
 static int
 lua_regexp_get_pattern (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct rspamd_lua_regexp *re = lua_check_regexp (L);
 
 	if (re && re->re && !IS_DESTROYED (re)) {
@@ -256,6 +422,7 @@ lua_regexp_get_pattern (lua_State *L)
 static int
 lua_regexp_set_limit (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct rspamd_lua_regexp *re = lua_check_regexp (L);
 	gint64 lim;
 
@@ -273,6 +440,52 @@ lua_regexp_set_limit (lua_State *L)
 	return 0;
 }
 
+/***
+ * @method re:set_max_hits(lim)
+ * Set maximum number of hits returned by a regexp
+ * @param {number} lim limit in hits count
+ * @return {number} old number of max hits
+ */
+static int
+lua_regexp_set_max_hits (lua_State *L)
+{
+	LUA_TRACE_POINT;
+	struct rspamd_lua_regexp *re = lua_check_regexp (L);
+	guint lim;
+
+	lim = luaL_checkinteger (L, 2);
+
+	if (re && re->re && !IS_DESTROYED (re)) {
+		lua_pushinteger (L, rspamd_regexp_set_maxhits (re->re, lim));
+	}
+	else {
+		lua_pushnil (L);
+	}
+
+	return 1;
+}
+
+/***
+ * @method re:get_max_hits(lim)
+ * Get maximum number of hits returned by a regexp
+ * @return {number} number of max hits
+ */
+static int
+lua_regexp_get_max_hits (lua_State *L)
+{
+	LUA_TRACE_POINT;
+	struct rspamd_lua_regexp *re = lua_check_regexp (L);
+
+	if (re && re->re && !IS_DESTROYED (re)) {
+		lua_pushinteger (L, rspamd_regexp_get_maxhits (re->re));
+	}
+	else {
+		lua_pushinteger (L, 1);
+	}
+
+	return 1;
+}
+
 /***
  * @method re:search(line[, raw[, capture]])
  * Search line in regular expression object. If line matches then this
@@ -303,6 +516,7 @@ lua_regexp_set_limit (lua_State *L)
 static int
 lua_regexp_search (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct rspamd_lua_regexp *re = lua_check_regexp (L);
 	const gchar *data = NULL;
 	struct rspamd_lua_text *t;
@@ -347,7 +561,7 @@ lua_regexp_search (lua_State *L)
 					captures)) {
 
 				if (capture) {
-					lua_newtable (L);
+					lua_createtable (L, captures->len, 0);
 
 					for (capn = 0; capn < captures->len; capn ++) {
 						cap = &g_array_index (captures, struct rspamd_re_capture,
@@ -396,6 +610,7 @@ lua_regexp_search (lua_State *L)
 static int
 lua_regexp_match (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct rspamd_lua_regexp *re = lua_check_regexp (L);
 	struct rspamd_lua_text *t;
 	const gchar *data = NULL;
@@ -453,6 +668,7 @@ lua_regexp_match (lua_State *L)
 static int
 lua_regexp_matchn (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct rspamd_lua_regexp *re = lua_check_regexp (L);
 	struct rspamd_lua_text *t;
 	const gchar *data = NULL, *start = NULL, *end = NULL;
@@ -499,7 +715,7 @@ lua_regexp_matchn (lua_State *L)
 				}
 			}
 
-			lua_pushnumber (L, matches);
+			lua_pushinteger (L, matches);
 
 			return 1;
 		}
@@ -524,6 +740,7 @@ lua_regexp_matchn (lua_State *L)
 static int
 lua_regexp_split (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct rspamd_lua_regexp *re = lua_check_regexp (L);
 	const gchar *data = NULL;
 	struct rspamd_lua_text *t;
@@ -538,6 +755,12 @@ lua_regexp_split (lua_State *L)
 		}
 		else if (lua_type (L, 2) == LUA_TUSERDATA) {
 			t = lua_check_text (L, 2);
+
+			if (t == NULL) {
+				lua_error (L);
+				return 0;
+			}
+
 			data = t->start;
 			len = t->len;
 			is_text = TRUE;
@@ -562,12 +785,15 @@ lua_regexp_split (lua_State *L)
 						rspamd_lua_setclass (L, "rspamd{text}", -1);
 						t->start = old_start;
 						t->len = start - old_start;
-						t->own = FALSE;
+						t->flags = 0;
 					}
 
 					lua_rawseti (L, -2, ++i);
 					matched = TRUE;
 				}
+				else if (start == end) {
+					break;
+				}
 				old_start = end;
 			}
 
@@ -584,7 +810,7 @@ lua_regexp_split (lua_State *L)
 					rspamd_lua_setclass (L, "rspamd{text}", -1);
 					t->start = end;
 					t->len = (data + len) - end;
-					t->own = FALSE;
+					t->flags = 0;
 				}
 
 				lua_rawseti (L, -2, ++i);
@@ -610,6 +836,7 @@ lua_regexp_split (lua_State *L)
 static gint
 lua_regexp_destroy (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct rspamd_lua_regexp *to_del = lua_check_regexp (L);
 
 	if (to_del) {
@@ -625,6 +852,7 @@ lua_regexp_destroy (lua_State *L)
 static gint
 lua_regexp_gc (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct rspamd_lua_regexp *to_del = lua_check_regexp (L);
 
 	if (to_del) {
@@ -632,7 +860,9 @@ lua_regexp_gc (lua_State *L)
 			rspamd_regexp_unref (to_del->re);
 		}
 
-		g_slice_free1 (sizeof (*to_del), to_del);
+		g_free (to_del->re_pattern);
+		g_free (to_del->module);
+		g_free (to_del);
 	}
 
 	return 0;
@@ -662,5 +892,10 @@ luaopen_regexp (lua_State * L)
 	luaL_register (L, NULL, regexplib_m);
 	rspamd_lua_add_preload (L, "rspamd_regexp", lua_load_regexp);
 
-	regexp_static_pool = rspamd_mempool_new (rspamd_mempool_suggest_size ());
+	if (regexp_static_pool == NULL) {
+		regexp_static_pool = rspamd_mempool_new (rspamd_mempool_suggest_size (),
+				"regexp_lua_pool");
+	}
+
+	lua_settop (L, 0);
 }
