@@ -1,30 +1,20 @@
-/* Copyright (c) 2010-2011, Vsevolod Stakhov
- * All rights reserved.
+/*-
+ * Copyright 2016 Vsevolod Stakhov
  *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions are met:
- *       * Redistributions of source code must retain the above copyright
- *         notice, this list of conditions and the following disclaimer.
- *       * Redistributions in binary form must reproduce the above copyright
- *         notice, this list of conditions and the following disclaimer in the
- *         documentation and/or other materials provided with the distribution.
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
  *
- * THIS SOFTWARE IS PROVIDED ''AS IS'' AND ANY
- * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
- * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
- * DISCLAIMED. IN NO EVENT SHALL AUTHOR BE LIABLE FOR ANY
- * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
- * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
- * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
- * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
- * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
- * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ *   http://www.apache.org/licenses/LICENSE-2.0
+ *
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
  */
-
 #include "config.h"
 #include "lua_common.h"
-#include "upstream.h"
-#include "cfg_file.h"
 
 
 /***
@@ -62,15 +52,19 @@ end
 /* Upstream list functions */
 LUA_FUNCTION_DEF (upstream_list, create);
 LUA_FUNCTION_DEF (upstream_list, destroy);
+LUA_FUNCTION_DEF (upstream_list, all_upstreams);
 LUA_FUNCTION_DEF (upstream_list, get_upstream_by_hash);
 LUA_FUNCTION_DEF (upstream_list, get_upstream_round_robin);
 LUA_FUNCTION_DEF (upstream_list, get_upstream_master_slave);
+LUA_FUNCTION_DEF (upstream_list, add_watcher);
 
 static const struct luaL_reg upstream_list_m[] = {
 
 	LUA_INTERFACE_DEF (upstream_list, get_upstream_by_hash),
 	LUA_INTERFACE_DEF (upstream_list, get_upstream_round_robin),
 	LUA_INTERFACE_DEF (upstream_list, get_upstream_master_slave),
+	LUA_INTERFACE_DEF (upstream_list, all_upstreams),
+	LUA_INTERFACE_DEF (upstream_list, add_watcher),
 	{"__tostring", rspamd_lua_class_tostring},
 	{"__gc", lua_upstream_list_destroy},
 	{NULL, NULL}
@@ -98,7 +92,7 @@ static const struct luaL_reg upstream_m[] = {
 static struct upstream *
 lua_check_upstream (lua_State * L)
 {
-	void *ud = luaL_checkudata (L, 1, "rspamd{upstream}");
+	void *ud = rspamd_lua_check_udata (L, 1, "rspamd{upstream}");
 
 	luaL_argcheck (L, ud != NULL, 1, "'upstream' expected");
 	return ud ? *((struct upstream **)ud) : NULL;
@@ -112,10 +106,11 @@ lua_check_upstream (lua_State * L)
 static gint
 lua_upstream_get_addr (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct upstream *up = lua_check_upstream (L);
 
 	if (up) {
-		rspamd_lua_ip_push (L, rspamd_upstream_addr (up));
+		rspamd_lua_ip_push (L, rspamd_upstream_addr_next (up));
 	}
 	else {
 		lua_pushnil (L);
@@ -131,10 +126,17 @@ lua_upstream_get_addr (lua_State *L)
 static gint
 lua_upstream_fail (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct upstream *up = lua_check_upstream (L);
+	gboolean fail_addr = FALSE;
 
 	if (up) {
-		rspamd_upstream_fail (up);
+
+		if (lua_isboolean (L, 2)) {
+			fail_addr = lua_toboolean (L, 2);
+		}
+
+		rspamd_upstream_fail (up, fail_addr);
 	}
 
 	return 0;
@@ -147,6 +149,7 @@ lua_upstream_fail (lua_State *L)
 static gint
 lua_upstream_ok (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct upstream *up = lua_check_upstream (L);
 
 	if (up) {
@@ -161,61 +164,79 @@ lua_upstream_ok (lua_State *L)
 static struct upstream_list *
 lua_check_upstream_list (lua_State * L)
 {
-	void *ud = luaL_checkudata (L, 1, "rspamd{upstream_list}");
+	void *ud = rspamd_lua_check_udata (L, 1, "rspamd{upstream_list}");
 
 	luaL_argcheck (L, ud != NULL, 1, "'upstream_list' expected");
 	return ud ? *((struct upstream_list **)ud) : NULL;
 }
 
 /***
- * @function upstream_list.create(def)
+ * @function upstream_list.create(cfg, def, [default_port])
  * Create new upstream list from its string definition in form `<upstream>,<upstream>;<upstream>`
+ * @param {rspamd_config} cfg configuration reference
  * @param {string} def upstream list definition
+ * @param {number} default_port default port for upstreams
  * @return {upstream_list} upstream list structure
  */
 static gint
 lua_upstream_list_create (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct upstream_list *new = NULL, **pnew;
+	struct rspamd_config *cfg = NULL;
 	const gchar *def;
-	char **tokens, **t;
 	guint default_port = 0;
+	gint top;
 
-	def = luaL_checkstring (L, 1);
-	if (def) {
-		if (lua_gettop (L) >= 2) {
-			default_port = luaL_checknumber (L, 2);
-		}
 
-		tokens = g_strsplit_set (def, ",;", 0);
-		if (!tokens || !tokens[0]) {
-			goto err;
-		}
-		t = tokens;
+	if (lua_type (L, 1) == LUA_TUSERDATA) {
+		cfg = lua_check_config (L, 1);
+		top = 2;
+	}
+	else {
+		top = 1;
+	}
 
-		new = rspamd_upstreams_create ();
-		while (*t != NULL) {
-			if (!rspamd_upstreams_add_upstream (new, *t, default_port, NULL)) {
-				goto err;
-			}
-			t ++;
+	if (lua_gettop (L) >= top + 1) {
+		default_port = luaL_checknumber (L, top + 1);
+	}
+
+	if (lua_type (L, top) == LUA_TSTRING) {
+		def = luaL_checkstring (L, top);
+
+		new = rspamd_upstreams_create (cfg ? cfg->ups_ctx : NULL);
+
+		if (rspamd_upstreams_parse_line (new, def, default_port, NULL)) {
+			pnew = lua_newuserdata (L, sizeof (struct upstream_list *));
+			rspamd_lua_setclass (L, "rspamd{upstream_list}", -1);
+			*pnew = new;
 		}
+		else {
+			rspamd_upstreams_destroy (new);
+			lua_pushnil (L);
+		}
+	}
+	else if (lua_type (L, top) == LUA_TTABLE) {
+		new = rspamd_upstreams_create (cfg ? cfg->ups_ctx : NULL);
 		pnew = lua_newuserdata (L, sizeof (struct upstream_list *));
 		rspamd_lua_setclass (L, "rspamd{upstream_list}", -1);
-
-		g_strfreev (tokens);
 		*pnew = new;
-	}
 
-	return 1;
-err:
-	if (tokens) {
-		g_strfreev (tokens);
+		lua_pushvalue (L, top);
+
+		for (lua_pushnil (L); lua_next (L, -2); lua_pop (L, 1)) {
+			def = lua_tostring (L, -1);
+
+			if (!def || !rspamd_upstreams_parse_line (new, def, default_port, NULL)) {
+				msg_warn ("cannot parse upstream %s", def);
+			}
+		}
+
+		lua_pop (L, 1);
 	}
-	if (new) {
-		rspamd_upstreams_destroy (new);
+	else {
+		return luaL_error (L, "invalid arguments");
 	}
-	lua_pushnil (L);
 
 	return 1;
 }
@@ -228,6 +249,7 @@ err:
 static gint
 lua_upstream_list_destroy (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct upstream_list *upl = lua_check_upstream_list (L);
 
 	rspamd_upstreams_destroy (upl);
@@ -244,6 +266,7 @@ lua_upstream_list_destroy (lua_State *L)
 static gint
 lua_upstream_list_get_upstream_by_hash (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct upstream_list *upl;
 	struct upstream *selected, **pselected;
 	const gchar *key;
@@ -269,7 +292,7 @@ lua_upstream_list_get_upstream_by_hash (lua_State *L)
 		}
 	}
 	else {
-		lua_pushnil (L);
+		return luaL_error (L, "invalid arguments");
 	}
 
 	return 1;
@@ -283,13 +306,14 @@ lua_upstream_list_get_upstream_by_hash (lua_State *L)
 static gint
 lua_upstream_list_get_upstream_round_robin (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct upstream_list *upl;
 	struct upstream *selected, **pselected;
 
 	upl = lua_check_upstream_list (L);
 	if (upl) {
 
-		selected = rspamd_upstream_get (upl, RSPAMD_UPSTREAM_ROUND_ROBIN);
+		selected = rspamd_upstream_get (upl, RSPAMD_UPSTREAM_ROUND_ROBIN, NULL, 0);
 		if (selected) {
 			pselected = lua_newuserdata (L, sizeof (struct upstream *));
 			rspamd_lua_setclass (L, "rspamd{upstream}", -1);
@@ -300,7 +324,7 @@ lua_upstream_list_get_upstream_round_robin (lua_State *L)
 		}
 	}
 	else {
-		lua_pushnil (L);
+		return luaL_error (L, "invalid arguments");
 	}
 
 	return 1;
@@ -314,13 +338,16 @@ lua_upstream_list_get_upstream_round_robin (lua_State *L)
 static gint
 lua_upstream_list_get_upstream_master_slave (lua_State *L)
 {
+	LUA_TRACE_POINT;
 	struct upstream_list *upl;
 	struct upstream *selected, **pselected;
 
 	upl = lua_check_upstream_list (L);
 	if (upl) {
 
-		selected = rspamd_upstream_get (upl, RSPAMD_UPSTREAM_MASTER_SLAVE);
+		selected = rspamd_upstream_get (upl, RSPAMD_UPSTREAM_MASTER_SLAVE,
+				NULL,
+				0);
 		if (selected) {
 			pselected = lua_newuserdata (L, sizeof (struct upstream *));
 			rspamd_lua_setclass (L, "rspamd{upstream}", -1);
@@ -331,12 +358,207 @@ lua_upstream_list_get_upstream_master_slave (lua_State *L)
 		}
 	}
 	else {
-		lua_pushnil (L);
+		return luaL_error (L, "invalid arguments");
 	}
 
 	return 1;
 }
 
+static void lua_upstream_inserter (struct upstream *up, guint idx, void *ud)
+{
+	struct upstream **pup;
+	lua_State *L = (lua_State *)ud;
+
+	pup = lua_newuserdata (L, sizeof (struct upstream *));
+	rspamd_lua_setclass (L, "rspamd{upstream}", -1);
+	*pup = up;
+
+	lua_rawseti (L, -2, idx + 1);
+}
+/***
+ * @method upstream_list:all_upstreams()
+ * Returns all upstreams for this list
+ * @return {table|upstream} all upstreams defined
+ */
+static gint
+lua_upstream_list_all_upstreams (lua_State *L)
+{
+	LUA_TRACE_POINT;
+	struct upstream_list *upl;
+
+	upl = lua_check_upstream_list (L);
+	if (upl) {
+		lua_createtable (L, rspamd_upstreams_count (upl), 0);
+		rspamd_upstreams_foreach (upl, lua_upstream_inserter, L);
+	}
+	else {
+		return luaL_error (L, "invalid arguments");
+	}
+
+	return 1;
+}
+
+static inline enum rspamd_upstreams_watch_event
+lua_str_to_upstream_flag (const gchar *str)
+{
+	enum rspamd_upstreams_watch_event fl = 0;
+
+	if (strcmp (str, "success") == 0) {
+		fl = RSPAMD_UPSTREAM_WATCH_SUCCESS;
+	}
+	else if (strcmp (str, "failure") == 0) {
+		fl = RSPAMD_UPSTREAM_WATCH_FAILURE;
+	}
+	else if (strcmp (str, "online") == 0) {
+		fl = RSPAMD_UPSTREAM_WATCH_ONLINE;
+	}
+	else if (strcmp (str, "offline") == 0) {
+		fl = RSPAMD_UPSTREAM_WATCH_OFFLINE;
+	}
+	else {
+		msg_err ("invalid flag: %s", str);
+	}
+
+	return fl;
+}
+
+static inline const gchar *
+lua_upstream_flag_to_str (enum rspamd_upstreams_watch_event fl)
+{
+	const gchar *res = "unknown";
+
+	/* Works with single flags, not combinations */
+	if (fl & RSPAMD_UPSTREAM_WATCH_SUCCESS) {
+		res = "success";
+	}
+	else if (fl & RSPAMD_UPSTREAM_WATCH_FAILURE) {
+		res = "failure";
+	}
+	else if (fl & RSPAMD_UPSTREAM_WATCH_ONLINE) {
+		res = "online";
+	}
+	else if (fl & RSPAMD_UPSTREAM_WATCH_OFFLINE) {
+		res = "offline";
+	}
+	else {
+		msg_err ("invalid flag: %d", fl);
+	}
+
+	return res;
+}
+
+struct rspamd_lua_upstream_watcher_cbdata {
+	lua_State *L;
+	gint cbref;
+	struct upstream_list *upl;
+};
+
+static void
+lua_upstream_watch_func (struct upstream *up,
+						 enum rspamd_upstreams_watch_event event,
+						 guint cur_errors,
+						 void *ud)
+{
+	struct rspamd_lua_upstream_watcher_cbdata *cdata =
+			(struct rspamd_lua_upstream_watcher_cbdata *)ud;
+	lua_State *L;
+	struct upstream **pup;
+	const gchar *what;
+	gint err_idx;
+
+	L = cdata->L;
+	what = lua_upstream_flag_to_str (event);
+	lua_pushcfunction (L, &rspamd_lua_traceback);
+	err_idx = lua_gettop (L);
+
+	lua_rawgeti (L, LUA_REGISTRYINDEX, cdata->cbref);
+	lua_pushstring (L, what);
+	pup = lua_newuserdata (L, sizeof (*pup));
+	*pup = up;
+	rspamd_lua_setclass (L, "rspamd{upstream}", -1);
+	lua_pushinteger (L, cur_errors);
+
+	if (lua_pcall (L, 3, 0, err_idx) != 0) {
+		GString *tb = lua_touserdata (L, -1);
+		msg_err ("cannot call watch function for upstream: %s", tb->str);
+		g_string_free (tb, TRUE);
+		lua_settop (L, 0);
+
+		return;
+	}
+
+	lua_settop (L, 0);
+}
+
+static void
+lua_upstream_watch_dtor (gpointer ud)
+{
+	struct rspamd_lua_upstream_watcher_cbdata *cdata =
+			(struct rspamd_lua_upstream_watcher_cbdata *)ud;
+
+	luaL_unref (cdata->L, LUA_REGISTRYINDEX, cdata->cbref);
+	g_free (cdata);
+}
+
+/***
+ * @method upstream_list:add_watcher(what, cb)
+ * Add new watcher to the upstream lists events (table or a string):
+ *   - `success` - called whenever upstream successfully used
+ *   - `failure` - called on upstream error
+ *   - `online` - called when upstream is being taken online from offline
+ *   - `offline` - called when upstream is being taken offline from online
+ * Callback is a function: function(what, upstream, cur_errors) ... end
+ * @example
+ups:add_watcher('success', function(what, up, cur_errors) ... end)
+ups:add_watcher({'online', 'offline'}, function(what, up, cur_errors) ... end)
+ * @return nothing
+ */
+static gint
+lua_upstream_list_add_watcher (lua_State *L)
+{
+	LUA_TRACE_POINT;
+	struct upstream_list *upl;
+
+	upl = lua_check_upstream_list (L);
+	if (upl &&
+		(lua_type (L, 2) == LUA_TTABLE ||  lua_type (L, 2) == LUA_TSTRING) &&
+		lua_type (L, 3) == LUA_TFUNCTION) {
+
+		enum rspamd_upstreams_watch_event flags = 0;
+		struct rspamd_lua_upstream_watcher_cbdata *cdata;
+
+		if (lua_type (L, 2) == LUA_TSTRING) {
+			flags = lua_str_to_upstream_flag (lua_tostring (L, 2));
+		}
+		else {
+			for (lua_pushnil (L); lua_next (L, -2); lua_pop (L, 1)) {
+				if (lua_isstring (L, -1)) {
+					flags |= lua_str_to_upstream_flag (lua_tostring (L, -1));
+				}
+				else {
+					lua_pop (L, 1);
+
+					return luaL_error (L, "invalid arguments");
+				}
+			}
+		}
+
+		cdata = g_malloc0 (sizeof (*cdata));
+		lua_pushvalue (L, 3); /* callback */
+		cdata->cbref = luaL_ref (L, LUA_REGISTRYINDEX);
+		cdata->L = L;
+		cdata->upl = upl;
+
+		rspamd_upstreams_add_watch_callback (upl, flags,
+				lua_upstream_watch_func, lua_upstream_watch_dtor, cdata);
+	}
+	else {
+		return luaL_error (L, "invalid arguments");
+	}
+
+	return 0;
+}
+
 static gint
 lua_load_upstream_list (lua_State * L)
 {
