@@ -1,29 +1,21 @@
-/* Copyright (c) 2011, Vsevolod Stakhov
- * All rights reserved.
+/*-
+ * Copyright 2016 Vsevolod Stakhov
  *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions are met:
- *       * Redistributions of source code must retain the above copyright
- *         notice, this list of conditions and the following disclaimer.
- *       * Redistributions in binary form must reproduce the above copyright
- *         notice, this list of conditions and the following disclaimer in the
- *         documentation and/or other materials provided with the distribution.
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
  *
- * THIS SOFTWARE IS PROVIDED ''AS IS'' AND ANY
- * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
- * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
- * DISCLAIMED. IN NO EVENT SHALL AUTHOR BE LIABLE FOR ANY
- * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
- * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
- * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
- * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
- * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
- * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ *   http://www.apache.org/licenses/LICENSE-2.0
+ *
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
  */
-
 #include "config.h"
 #include "tests.h"
-#include "main.h"
+#include "rspamd.h"
 #include "dkim.h"
 
 static const gchar test_dkim_sig[] = "v=1; a=rsa-sha256; c=relaxed/relaxed; "
@@ -42,7 +34,7 @@ test_key_handler (rspamd_dkim_key_t *key, gsize keylen, rspamd_dkim_context_t *c
 	struct rspamd_async_session *s = ud;
 	g_assert (key != NULL);
 
-	destroy_session (s);
+	rspamd_session_destroy (s);
 }
 
 static gboolean
@@ -60,6 +52,7 @@ session_fin (gpointer unused)
 void
 rspamd_dkim_test_func ()
 {
+#if 0
 	rspamd_dkim_context_t *ctx;
 	rspamd_dkim_key_t *key;
 	rspamd_mempool_t *pool;
@@ -85,9 +78,10 @@ rspamd_dkim_test_func ()
 	g_assert (ctx != NULL);
 
 	/* Key part */
-	s = new_async_session (pool, session_fin, NULL, NULL, NULL);
+	s = rspamd_session_create (pool, session_fin, NULL, NULL, NULL);
 
 	g_assert (rspamd_get_dkim_key (ctx, resolver, s, test_key_handler, s));
 
 	event_base_loop (base, 0);
+#endif
 }
