@@ -1,32 +1,22 @@
-/*
- * Copyright (c) 2014, Vsevolod Stakhov
+/*-
+ * Copyright 2016 Vsevolod Stakhov
  *
- * All rights reserved.
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
  *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions are met:
- *	 * Redistributions of source code must retain the above copyright
- *	   notice, this list of conditions and the following disclaimer.
- *	 * Redistributions in binary form must reproduce the above copyright
- *	   notice, this list of conditions and the following disclaimer in the
- *	   documentation and/or other materials provided with the distribution.
+ *   http://www.apache.org/licenses/LICENSE-2.0
  *
- * THIS SOFTWARE IS PROVIDED BY AUTHOR ''AS IS'' AND ANY
- * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
- * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
- * DISCLAIMED. IN NO EVENT SHALL AUTHOR BE LIABLE FOR ANY
- * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
- * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
- * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
- * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
- * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
- * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
  */
-
 #include "config.h"
-#include "main.h"
-#include "upstream.h"
+#include "rspamd.h"
 #include "ottery.h"
+#include <math.h>
 
 const char *test_upstream_list = "microsoft.com:443:1,google.com:80:2,kernel.org:443:3";
 const char *new_upstream_list = "freebsd.org:80";
@@ -39,7 +29,7 @@ rspamd_upstream_test_method (struct upstream_list *ls,
 	struct upstream *up;
 
 	if (rot != RSPAMD_UPSTREAM_HASHED) {
-		up = rspamd_upstream_get (ls, rot);
+		up = rspamd_upstream_get (ls, rot, NULL, 0);
 		g_assert (up != NULL);
 		g_assert (strcmp (rspamd_upstream_name (up), expected) == 0);
 	}
@@ -74,21 +64,58 @@ rspamd_upstream_test_func (void)
 	struct timeval tv;
 	rspamd_inet_addr_t *addr, *next_addr, *paddr;
 
-	cfg = (struct rspamd_config *)g_malloc (sizeof (struct rspamd_config));
-	bzero (cfg, sizeof (struct rspamd_config));
-	cfg->cfg_pool = rspamd_mempool_new (rspamd_mempool_suggest_size ());
+	cfg = rspamd_config_new (RSPAMD_CONFIG_INIT_SKIP_LUA);
 	cfg->dns_retransmits = 2;
 	cfg->dns_timeout = 0.5;
 	cfg->upstream_max_errors = 1;
 	cfg->upstream_revive_time = 0.5;
 	cfg->upstream_error_time = 2;
 
-	resolver = dns_resolver_init (NULL, ev_base, cfg);
+	resolver = rspamd_dns_resolver_init (NULL, ev_base, cfg);
+	rspamd_upstreams_library_config (cfg, cfg->ups_ctx, ev_base, resolver->r);
 
-	rspamd_upstreams_library_init (resolver->r, ev_base);
-	rspamd_upstreams_library_config (cfg);
+	/*
+	 * Test v4/v6 priorities
+	 */
+	nls = rspamd_upstreams_create (cfg->ups_ctx);
+	g_assert (rspamd_upstreams_add_upstream (nls, "127.0.0.1", 0,
+			RSPAMD_UPSTREAM_PARSE_DEFAULT,
+			NULL));
+	up = rspamd_upstream_get (nls, RSPAMD_UPSTREAM_RANDOM, NULL, 0);
+	rspamd_parse_inet_address (&paddr, "127.0.0.2", 0);
+	g_assert (rspamd_upstream_add_addr (up, paddr));
+	rspamd_parse_inet_address (&paddr, "::1", 0);
+	g_assert (rspamd_upstream_add_addr (up, paddr));
+	/* Rewind to start */
+	addr = rspamd_upstream_addr_next (up);
+	addr = rspamd_upstream_addr_next (up);
+	/* cur should be zero here */
+	addr = rspamd_upstream_addr_next (up);
+	next_addr = rspamd_upstream_addr_next (up);
+	g_assert (rspamd_inet_address_get_af (addr) == AF_INET);
+	g_assert (rspamd_inet_address_get_af (next_addr) == AF_INET);
+	next_addr = rspamd_upstream_addr_next (up);
+	g_assert (rspamd_inet_address_get_af (next_addr) == AF_INET6);
+	next_addr = rspamd_upstream_addr_next (up);
+	g_assert (rspamd_inet_address_get_af (next_addr) == AF_INET);
+	next_addr = rspamd_upstream_addr_next (up);
+	g_assert (rspamd_inet_address_get_af (next_addr) == AF_INET);
+	next_addr = rspamd_upstream_addr_next (up);
+	g_assert (rspamd_inet_address_get_af (next_addr) == AF_INET6);
+	/* Test errors with IPv6 */
+	rspamd_upstream_fail (up, TRUE);
+	/* Now we should have merely IPv4 addresses in rotation */
+	addr = rspamd_upstream_addr_next (up);
+	for (i = 0; i < 256; i++) {
+		next_addr = rspamd_upstream_addr_next (up);
+		g_assert (rspamd_inet_address_get_af (addr) == AF_INET);
+		g_assert (rspamd_inet_address_get_af (next_addr) == AF_INET);
+		g_assert (rspamd_inet_address_compare (addr, next_addr, FALSE) != 0);
+		addr = next_addr;
+	}
+	rspamd_upstreams_destroy (nls);
 
-	ls = rspamd_upstreams_create ();
+	ls = rspamd_upstreams_create (cfg->ups_ctx);
 	g_assert (rspamd_upstreams_parse_line (ls, test_upstream_list, 443, NULL));
 	g_assert (rspamd_upstreams_count (ls) == 3);
 
@@ -105,7 +132,7 @@ rspamd_upstream_test_func (void)
 	rspamd_upstream_test_method (ls, RSPAMD_UPSTREAM_ROUND_ROBIN, "microsoft.com");
 
 	/* Test stable hashing */
-	nls = rspamd_upstreams_create ();
+	nls = rspamd_upstreams_create (cfg->ups_ctx);
 	g_assert (rspamd_upstreams_parse_line (nls, test_upstream_list, 443, NULL));
 	g_assert (rspamd_upstreams_parse_line (nls, new_upstream_list, 443, NULL));
 	for (i = 0; i < assumptions; i ++) {
@@ -132,33 +159,14 @@ rspamd_upstream_test_func (void)
 
 	rspamd_upstreams_destroy (nls);
 
-	/*
-	 * Test v4/v6 priorities
-	 */
-	nls = rspamd_upstreams_create ();
-	g_assert (rspamd_upstreams_add_upstream (nls, "127.0.0.1", 0, NULL));
-	up = rspamd_upstream_get (nls, RSPAMD_UPSTREAM_RANDOM);
-	rspamd_parse_inet_address(&paddr, "127.0.0.2");
-	g_assert (rspamd_upstream_add_addr (up, paddr));
-	rspamd_parse_inet_address(&paddr, "::1");
-	g_assert (rspamd_upstream_add_addr (up, paddr));
-	addr = rspamd_upstream_addr (up);
-	for (i = 0; i < 256; i ++) {
-		next_addr = rspamd_upstream_addr (up);
-		g_assert (rspamd_inet_address_get_af (addr) == AF_INET);
-		g_assert (rspamd_inet_address_get_af (next_addr) == AF_INET);
-		g_assert (rspamd_inet_address_compare (addr, next_addr) != 0);
-		addr = next_addr;
-	}
-	rspamd_upstreams_destroy (nls);
 
 	/* Upstream fail test */
 	evtimer_set (&ev, rspamd_upstream_timeout_handler, resolver);
 	event_base_set (ev_base, &ev);
 
-	up = rspamd_upstream_get (ls, RSPAMD_UPSTREAM_MASTER_SLAVE);
+	up = rspamd_upstream_get (ls, RSPAMD_UPSTREAM_MASTER_SLAVE, NULL, 0);
 	for (i = 0; i < 100; i ++) {
-		rspamd_upstream_fail (up);
+		rspamd_upstream_fail (up, TRUE);
 	}
 	g_assert (rspamd_upstreams_alive (ls) == 2);
 
@@ -169,6 +177,6 @@ rspamd_upstream_test_func (void)
 	event_base_loop (ev_base, 0);
 	g_assert (rspamd_upstreams_alive (ls) == 3);
 
-	g_free (cfg);
 	rspamd_upstreams_destroy (ls);
+	REF_RELEASE (cfg);
 }
